"""Tests for the aldakit parser."""

import pytest
from aldakit.parser import parse
from aldakit.ast_nodes import (
    RootNode,
    PartNode,
    EventSequenceNode,
    NoteNode,
    RestNode,
    ChordNode,
    NoteLengthNode,
    NoteLengthMsNode,
    NoteLengthSecondsNode,
    BarlineNode,
    OctaveSetNode,
    OctaveUpNode,
    OctaveDownNode,
    LispListNode,
    LispSymbolNode,
    LispNumberNode,
    VariableDefinitionNode,
    VariableReferenceNode,
    MarkerNode,
    AtMarkerNode,
    VoiceGroupNode,
    CramNode,
    RepeatNode,
    OnRepetitionsNode,
    BracketedSequenceNode,
)
from aldakit.errors import AldaSyntaxError


class TestBasicParsing:
    """Test basic parsing functionality."""

    def test_empty_input(self):
        ast = parse("")
        assert isinstance(ast, RootNode)
        assert len(ast.children) == 0

    def test_single_note(self):
        ast = parse("c")
        assert len(ast.children) == 1
        seq = ast.children[0]
        assert isinstance(seq, EventSequenceNode)
        assert len(seq.events) == 1
        note = seq.events[0]
        assert isinstance(note, NoteNode)
        assert note.letter == "c"

    def test_rest(self):
        ast = parse("r")
        seq = ast.children[0]
        rest = seq.events[0]
        assert isinstance(rest, RestNode)


class TestNotes:
    """Test note parsing."""

    def test_note_with_accidentals(self):
        ast = parse("c+")
        note = ast.children[0].events[0]
        assert note.letter == "c"
        assert note.accidentals == ["+"]

    def test_note_with_multiple_accidentals(self):
        ast = parse("c++")
        note = ast.children[0].events[0]
        assert note.accidentals == ["+", "+"]

    def test_note_with_flat(self):
        ast = parse("b-")
        note = ast.children[0].events[0]
        assert note.letter == "b"
        assert note.accidentals == ["-"]

    def test_note_with_natural(self):
        ast = parse("f_")
        note = ast.children[0].events[0]
        assert note.accidentals == ["_"]

    def test_note_with_duration(self):
        ast = parse("c4")
        note = ast.children[0].events[0]
        assert note.letter == "c"
        assert note.duration is not None
        assert isinstance(note.duration.components[0], NoteLengthNode)
        assert note.duration.components[0].denominator == 4

    def test_note_with_dotted_duration(self):
        ast = parse("c4.")
        note = ast.children[0].events[0]
        length = note.duration.components[0]
        assert length.denominator == 4
        assert length.dots == 1

    def test_note_with_double_dot(self):
        ast = parse("c4..")
        note = ast.children[0].events[0]
        length = note.duration.components[0]
        assert length.dots == 2

    def test_note_with_ms_duration(self):
        ast = parse("c500ms")
        note = ast.children[0].events[0]
        ms = note.duration.components[0]
        assert isinstance(ms, NoteLengthMsNode)
        assert ms.ms == 500.0

    def test_note_with_seconds_duration(self):
        ast = parse("c2s")
        note = ast.children[0].events[0]
        sec = note.duration.components[0]
        assert isinstance(sec, NoteLengthSecondsNode)
        assert sec.seconds == 2.0

    def test_slurred_note(self):
        ast = parse("c~d")
        seq = ast.children[0]
        assert seq.events[0].slurred is True
        assert seq.events[1].slurred is False


class TestChords:
    """Test chord parsing."""

    def test_simple_chord(self):
        ast = parse("c/e/g")
        chord = ast.children[0].events[0]
        assert isinstance(chord, ChordNode)
        assert len(chord.notes) == 3
        notes = [n for n in chord.notes if isinstance(n, NoteNode)]
        assert [n.letter for n in notes] == ["c", "e", "g"]

    def test_chord_with_duration(self):
        ast = parse("c1/e/g")
        chord = ast.children[0].events[0]
        first_note = chord.notes[0]
        assert first_note.duration is not None
        assert first_note.duration.components[0].denominator == 1

    def test_chord_with_octave_changes(self):
        ast = parse("c/>e/g")
        chord = ast.children[0].events[0]
        # Should have: NoteNode, OctaveUpNode, NoteNode, NoteNode
        assert isinstance(chord.notes[0], NoteNode)
        assert isinstance(chord.notes[1], OctaveUpNode)
        assert isinstance(chord.notes[2], NoteNode)


class TestOctaves:
    """Test octave control parsing."""

    def test_octave_set(self):
        ast = parse("o4 c")
        seq = ast.children[0]
        assert isinstance(seq.events[0], OctaveSetNode)
        assert seq.events[0].octave == 4

    def test_octave_up(self):
        ast = parse("> c")
        seq = ast.children[0]
        assert isinstance(seq.events[0], OctaveUpNode)

    def test_octave_down(self):
        ast = parse("< c")
        seq = ast.children[0]
        assert isinstance(seq.events[0], OctaveDownNode)


class TestParts:
    """Test part declaration parsing."""

    def test_simple_part(self):
        ast = parse("piano: c d e")
        assert len(ast.children) == 1
        part = ast.children[0]
        assert isinstance(part, PartNode)
        assert part.declaration.names == ["piano"]
        assert len(part.events.events) == 3

    def test_part_with_alias(self):
        ast = parse('violin "v1": c d e')
        part = ast.children[0]
        assert part.declaration.names == ["violin"]
        assert part.declaration.alias == "v1"

    def test_multi_instrument_part(self):
        ast = parse("violin/viola: c d e")
        part = ast.children[0]
        assert part.declaration.names == ["violin", "viola"]

    def test_multiple_parts(self):
        ast = parse("piano: c d e\nviolin: f g a")
        assert len(ast.children) == 2
        assert ast.children[0].declaration.names == ["piano"]
        assert ast.children[1].declaration.names == ["violin"]


class TestBarlines:
    """Test barline parsing."""

    def test_barline(self):
        ast = parse("c d | e f")
        seq = ast.children[0]
        events = seq.events
        assert isinstance(events[2], BarlineNode)


class TestSExpressions:
    """Test S-expression parsing."""

    def test_tempo(self):
        ast = parse("(tempo 120)")
        sexp = ast.children[0].events[0]
        assert isinstance(sexp, LispListNode)
        assert len(sexp.elements) == 2
        assert isinstance(sexp.elements[0], LispSymbolNode)
        assert sexp.elements[0].name == "tempo"
        assert isinstance(sexp.elements[1], LispNumberNode)
        assert sexp.elements[1].value == 120

    def test_volume(self):
        ast = parse("(vol 80)")
        sexp = ast.children[0].events[0]
        assert sexp.elements[0].name == "vol"
        assert sexp.elements[1].value == 80

    def test_nested_sexp(self):
        ast = parse("(foo (bar 42))")
        sexp = ast.children[0].events[0]
        assert isinstance(sexp.elements[1], LispListNode)

    def test_sexp_in_sequence(self):
        ast = parse("(tempo 120) c d e")
        seq = ast.children[0]
        assert isinstance(seq.events[0], LispListNode)
        assert isinstance(seq.events[1], NoteNode)


class TestComplexExamples:
    """Test complete Alda snippets."""

    def test_melody_with_attributes(self):
        source = """
        piano:
          (tempo 120)
          o4 c8 d e f | g4 a b > c
        """
        ast = parse(source)
        part = ast.children[0]
        assert isinstance(part, PartNode)
        events = part.events.events
        # Should have: tempo sexp, octave set, notes, barline, more notes, octave up, note
        assert isinstance(events[0], LispListNode)
        assert isinstance(events[1], OctaveSetNode)

    def test_chord_progression(self):
        source = "c1/e/g | f/a/c | g/b/d | c/e/g"
        ast = parse(source)
        seq = ast.children[0]
        chords = [e for e in seq.events if isinstance(e, ChordNode)]
        assert len(chords) == 4

    def test_tied_notes(self):
        source = "c1~1"
        ast = parse(source)
        seq = ast.children[0]
        # The tie connects durations, so we should have one note with tied duration
        note = seq.events[0]
        assert isinstance(note, NoteNode)
        assert len(note.duration.components) == 2

    def test_multiline_part(self):
        source = """
        piano:
          c d e f
          g a b > c
        """
        ast = parse(source)
        part = ast.children[0]
        assert len(part.events.events) > 0


class TestErrors:
    """Test error handling."""

    def test_name_without_colon_is_not_part(self):
        # 'piano' without colon is not a part declaration
        # It gets skipped (would be a variable reference in full implementation)
        ast = parse("piano c d e")
        # Should parse as implicit events (notes only, piano is skipped)
        seq = ast.children[0]
        notes = [e for e in seq.events if isinstance(e, NoteNode)]
        assert len(notes) == 3

    def test_unclosed_sexp(self):
        with pytest.raises(AldaSyntaxError):
            parse("(tempo 120")


class TestMultiInstrumentParts:
    """Test multi-instrument part declarations with tricky names."""

    def test_instruments_starting_with_note_letters(self):
        # cello, clarinet, flute, etc. start with note letters
        ast = parse('violin/viola/cello "strings": g1')
        part = ast.children[0]
        assert isinstance(part, PartNode)
        assert part.declaration.names == ["violin", "viola", "cello"]
        assert part.declaration.alias == "strings"

    def test_instrument_cello_alone(self):
        ast = parse("cello: c d e")
        part = ast.children[0]
        assert part.declaration.names == ["cello"]

    def test_instrument_flute_alone(self):
        ast = parse("flute: f g a")
        part = ast.children[0]
        assert part.declaration.names == ["flute"]


class TestRest:
    """Test rest parsing."""

    def test_rest_with_duration(self):
        ast = parse("r4")
        rest = ast.children[0].events[0]
        assert isinstance(rest, RestNode)
        assert rest.duration is not None
        assert rest.duration.components[0].denominator == 4

    def test_rest_with_dotted_duration(self):
        ast = parse("r4.")
        rest = ast.children[0].events[0]
        assert rest.duration.components[0].dots == 1


class TestVariables:
    """Test parsing of variables."""

    def test_variable_definition(self):
        ast = parse("myMotif = c d e")
        seq = ast.children[0]
        var_def = seq.events[0]
        assert isinstance(var_def, VariableDefinitionNode)
        assert var_def.name == "myMotif"
        assert len(var_def.events.events) == 3

    def test_variable_reference(self):
        ast = parse("myMotif")
        seq = ast.children[0]
        var_ref = seq.events[0]
        assert isinstance(var_ref, VariableReferenceNode)
        assert var_ref.name == "myMotif"

    def test_variable_definition_and_reference(self):
        ast = parse("theme = c d e\ntheme")
        seq = ast.children[0]
        assert isinstance(seq.events[0], VariableDefinitionNode)
        assert isinstance(seq.events[1], VariableReferenceNode)


class TestMarkers:
    """Test parsing of markers."""

    def test_marker(self):
        ast = parse("%verse")
        seq = ast.children[0]
        marker = seq.events[0]
        assert isinstance(marker, MarkerNode)
        assert marker.name == "verse"

    def test_at_marker(self):
        ast = parse("@chorus")
        seq = ast.children[0]
        at_marker = seq.events[0]
        assert isinstance(at_marker, AtMarkerNode)
        assert at_marker.name == "chorus"

    def test_marker_in_sequence(self):
        ast = parse("c d %bridge e f @bridge")
        seq = ast.children[0]
        events = seq.events
        assert isinstance(events[2], MarkerNode)
        assert isinstance(events[5], AtMarkerNode)


class TestVoices:
    """Test parsing of voice groups."""

    def test_single_voice(self):
        ast = parse("V1: c d e V0:")
        seq = ast.children[0]
        voice_group = seq.events[0]
        assert isinstance(voice_group, VoiceGroupNode)
        assert len(voice_group.voices) == 1
        assert voice_group.voices[0].number == 1

    def test_multiple_voices(self):
        ast = parse("V1: c d e V2: f g a V0:")
        seq = ast.children[0]
        voice_group = seq.events[0]
        assert isinstance(voice_group, VoiceGroupNode)
        assert len(voice_group.voices) == 2
        assert voice_group.voices[0].number == 1
        assert voice_group.voices[1].number == 2

    def test_voice_content(self):
        ast = parse("V1: c d e V0:")
        seq = ast.children[0]
        voice_group = seq.events[0]
        voice1 = voice_group.voices[0]
        assert len(voice1.events.events) == 3
        assert all(isinstance(e, NoteNode) for e in voice1.events.events)


class TestCram:
    """Test parsing of cram expressions."""

    def test_simple_cram(self):
        ast = parse("{c d e}")
        seq = ast.children[0]
        cram = seq.events[0]
        assert isinstance(cram, CramNode)
        assert len(cram.events.events) == 3

    def test_cram_with_duration(self):
        ast = parse("{c d e}2")
        seq = ast.children[0]
        cram = seq.events[0]
        assert isinstance(cram, CramNode)
        assert cram.duration is not None
        assert cram.duration.components[0].denominator == 2

    def test_nested_cram(self):
        ast = parse("{c {d e} f}")
        seq = ast.children[0]
        cram = seq.events[0]
        assert isinstance(cram, CramNode)
        # Should have 3 events: note, cram, note
        assert isinstance(cram.events.events[1], CramNode)


class TestBracketedSequences:
    """Test parsing of bracketed sequences."""

    def test_simple_sequence(self):
        ast = parse("[c d e]")
        seq = ast.children[0]
        bracketed = seq.events[0]
        assert isinstance(bracketed, BracketedSequenceNode)
        assert len(bracketed.events.events) == 3

    def test_sequence_with_repeat(self):
        ast = parse("[c d e]*4")
        seq = ast.children[0]
        repeat = seq.events[0]
        assert isinstance(repeat, RepeatNode)
        assert repeat.times == 4
        assert isinstance(repeat.event, BracketedSequenceNode)


class TestRepeats:
    """Test parsing of repeat expressions."""

    def test_repeat_note(self):
        ast = parse("c*4")
        seq = ast.children[0]
        repeat = seq.events[0]
        assert isinstance(repeat, RepeatNode)
        assert repeat.times == 4
        assert isinstance(repeat.event, NoteNode)

    def test_repeat_sequence(self):
        ast = parse("[c d e]*3")
        seq = ast.children[0]
        repeat = seq.events[0]
        assert isinstance(repeat, RepeatNode)
        assert repeat.times == 3


class TestOnRepetitions:
    """Test parsing of on-repetition expressions."""

    def test_simple_on_repetition(self):
        ast = parse("c'1")
        seq = ast.children[0]
        on_rep = seq.events[0]
        assert isinstance(on_rep, OnRepetitionsNode)
        assert len(on_rep.ranges) == 1
        assert on_rep.ranges[0].first == 1
        assert on_rep.ranges[0].last is None

    def test_range_on_repetition(self):
        ast = parse("c'1-3")
        seq = ast.children[0]
        on_rep = seq.events[0]
        assert isinstance(on_rep, OnRepetitionsNode)
        assert on_rep.ranges[0].first == 1
        assert on_rep.ranges[0].last == 3

    def test_multiple_ranges(self):
        ast = parse("c'1-3,5")
        seq = ast.children[0]
        on_rep = seq.events[0]
        assert isinstance(on_rep, OnRepetitionsNode)
        assert len(on_rep.ranges) == 2
        assert on_rep.ranges[0].first == 1
        assert on_rep.ranges[0].last == 3
        assert on_rep.ranges[1].first == 5


class TestComplexExpressions:
    """Test complex combinations of features."""

    def test_variable_with_cram(self):
        ast = parse("triplet = {c d e}4")
        seq = ast.children[0]
        var_def = seq.events[0]
        assert isinstance(var_def, VariableDefinitionNode)
        cram = var_def.events.events[0]
        assert isinstance(cram, CramNode)

    def test_repeated_variable_reference(self):
        ast = parse("theme = c d e\ntheme*4")
        seq = ast.children[0]
        repeat = seq.events[1]
        assert isinstance(repeat, RepeatNode)
        assert isinstance(repeat.event, VariableReferenceNode)

    def test_voices_with_markers(self):
        ast = parse("V1: %melody c d e V2: %harmony f g a V0:")
        seq = ast.children[0]
        voice_group = seq.events[0]
        # Each voice should have a marker followed by notes
        v1_events = voice_group.voices[0].events.events
        assert isinstance(v1_events[0], MarkerNode)
        assert v1_events[0].name == "melody"

    def test_bracketed_sequence_with_on_repetitions(self):
        ast = parse("[c'1 d'2 e]*4")
        seq = ast.children[0]
        repeat = seq.events[0]
        assert isinstance(repeat, RepeatNode)
        bracketed = repeat.event
        assert isinstance(bracketed, BracketedSequenceNode)
        # First event should be on-repetition
        assert isinstance(bracketed.events.events[0], OnRepetitionsNode)



"""Tests for scales, chords, and advanced compose elements (cram, voices, variables, markers)."""

import pytest

from aldakit.compose import (
    # Scales
    scale,
    scale_notes,
    scale_degree,
    mode,
    relative_minor,
    relative_major,
    parallel_minor,
    parallel_major,
    transpose_scale,
    interval_name,
    list_scales,
    build_chord,
    major,
    minor,
    dim,
    aug,
    sus2,
    sus4,
    maj7,
    min7,
    dom7,
    dim7,
    half_dim7,
    aug7,
    dom9,
    maj9,
    min9,
    add9,
    power,
    arpeggiate,
    invert_chord,
    voicing,
    list_chord_types,
    Cram,
    Voice,
    Variable,
    VariableRef,
    Marker,
    AtMarker,
    # Core functions
    note,
    cram,
    voice,
    voice_group,
    var,
    var_ref,
    marker,
    at_marker,
    seq,
)


# =============================================================================
# Scale Tests
# =============================================================================


class TestScale:
    """Tests for the scale function."""

    def test_c_major_scale(self):
        """C major scale should return correct pitch names."""
        result = scale("c", "major")
        assert result == ["c", "d", "e", "f", "g", "a", "b"]

    def test_a_minor_scale(self):
        """A minor scale (natural minor / aeolian)."""
        result = scale("a", "minor")
        assert result == ["a", "b", "c", "d", "e", "f", "g"]

    def test_c_pentatonic_scale(self):
        """C major pentatonic scale."""
        result = scale("c", "pentatonic")
        assert result == ["c", "d", "e", "g", "a"]

    def test_a_blues_scale(self):
        """A blues scale."""
        result = scale("a", "blues")
        # a, c, d, d#/eb, e, g
        assert result == ["a", "c", "d", "d+", "e", "g"]

    def test_f_lydian_mode(self):
        """F lydian mode."""
        result = scale("f", "lydian")
        assert result == ["f", "g", "a", "b", "c", "d", "e"]

    def test_d_dorian_mode(self):
        """D dorian mode."""
        result = scale("d", "dorian")
        assert result == ["d", "e", "f", "g", "a", "b", "c"]

    def test_chromatic_scale(self):
        """Chromatic scale has 12 notes."""
        result = scale("c", "chromatic")
        assert len(result) == 12

    def test_unknown_scale_raises(self):
        """Unknown scale type should raise ValueError."""
        with pytest.raises(ValueError, match="Unknown scale type"):
            scale("c", "nonexistent")

    def test_invalid_root_raises(self):
        """Invalid root note should raise ValueError."""
        with pytest.raises(ValueError, match="Invalid root note"):
            scale("x", "major")


class TestScaleNotes:
    """Tests for scale_notes function (returns Seq)."""

    def test_returns_seq(self):
        """scale_notes should return a Seq object."""
        result = scale_notes("c", "major")
        assert hasattr(result, "elements")
        assert len(result.elements) == 7

    def test_notes_have_correct_pitches(self):
        """Notes should have correct pitch names."""
        result = scale_notes("c", "major")
        pitches = [n.pitch for n in result.elements]
        assert pitches == ["c", "d", "e", "f", "g", "a", "b"]

    def test_with_duration(self):
        """Duration should be applied to all notes."""
        result = scale_notes("c", "major", duration=8)
        for n in result.elements:
            assert n.duration == 8

    def test_ascending_false(self):
        """Descending scale when ascending=False."""
        result = scale_notes("c", "major", ascending=False)
        pitches = [n.pitch for n in result.elements]
        assert pitches == ["b", "a", "g", "f", "e", "d", "c"]


class TestScaleDegree:
    """Tests for scale_degree function."""

    def test_first_degree_is_root(self):
        """Scale degree 1 should be the root."""
        pitch, acc, oct = scale_degree("c", "major", 1)
        assert pitch == "c"
        assert acc is None
        assert oct == 4

    def test_fifth_degree(self):
        """Scale degree 5 in C major is G."""
        pitch, acc, oct = scale_degree("c", "major", 5)
        assert pitch == "g"
        assert acc is None
        assert oct == 4

    def test_octave_wrapping(self):
        """Scale degree 8 should be root up an octave."""
        pitch, acc, oct = scale_degree("c", "major", 8)
        assert pitch == "c"
        assert oct == 5

    def test_degree_15_is_two_octaves(self):
        """Scale degree 15 should be root up two octaves."""
        pitch, acc, oct = scale_degree("c", "major", 15)
        assert pitch == "c"
        assert oct == 6

    def test_invalid_degree_raises(self):
        """Degree less than 1 should raise ValueError."""
        with pytest.raises(ValueError, match="Scale degree must be >= 1"):
            scale_degree("c", "major", 0)


class TestModeFunction:
    """Tests for mode function (alias for scale)."""

    def test_mode_is_alias_for_scale(self):
        """mode() should return same as scale()."""
        assert mode("d", "dorian") == scale("d", "dorian")


class TestRelativeMinorMajor:
    """Tests for relative minor/major functions."""

    def test_relative_minor_of_c(self):
        """Relative minor of C major is A minor."""
        assert relative_minor("c") == "a"

    def test_relative_minor_of_g(self):
        """Relative minor of G major is E minor."""
        assert relative_minor("g") == "e"

    def test_relative_major_of_a(self):
        """Relative major of A minor is C major."""
        assert relative_major("a") == "c"

    def test_relative_major_of_e(self):
        """Relative major of E minor is G major."""
        assert relative_major("e") == "g"

    def test_round_trip(self):
        """Relative major of relative minor should return original."""
        assert relative_major(relative_minor("c")) == "c"


class TestParallelMinorMajor:
    """Tests for parallel minor/major functions."""

    def test_parallel_minor_same_root(self):
        """Parallel minor has same root."""
        assert parallel_minor("c") == "c"

    def test_parallel_major_same_root(self):
        """Parallel major has same root."""
        assert parallel_major("a") == "a"


class TestTransposeScale:
    """Tests for transpose_scale function."""

    def test_transpose_up_fifth(self):
        """Transposing C D E up a fifth (7 semitones)."""
        result = transpose_scale(["c", "d", "e"], 7)
        assert result == ["g", "a", "b"]

    def test_transpose_up_fourth(self):
        """Transposing C D E up a fourth (5 semitones)."""
        result = transpose_scale(["c", "d", "e"], 5)
        assert result == ["f", "g", "a"]

    def test_transpose_with_accidentals(self):
        """Transposing notes with accidentals."""
        result = transpose_scale(["c+", "d+"], 2)
        assert result == ["d+", "f"]


class TestIntervalName:
    """Tests for interval_name function."""

    def test_unison(self):
        assert interval_name(0) == "unison"

    def test_perfect_fifth(self):
        assert interval_name(7) == "perfect fifth"

    def test_major_third(self):
        assert interval_name(4) == "major third"

    def test_octave(self):
        assert interval_name(12) == "octave"


class TestListScales:
    """Tests for list_scales function."""

    def test_returns_list(self):
        result = list_scales()
        assert isinstance(result, list)

    def test_contains_common_scales(self):
        result = list_scales()
        assert "major" in result
        assert "minor" in result
        assert "pentatonic" in result


# =============================================================================
# Chord Tests
# =============================================================================


class TestBuildChord:
    """Tests for build_chord function."""

    def test_c_major_chord(self):
        """C major chord should have C E G."""
        chord = build_chord("c", "major")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["c", "e", "g"]

    def test_a_minor_chord(self):
        """A minor chord should have A C E."""
        chord = build_chord("a", "minor")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["a", "c", "e"]

    def test_chord_with_duration(self):
        """Chord should accept duration parameter."""
        chord = build_chord("c", "major", duration=2)
        assert chord.duration == 2

    def test_chord_inversion(self):
        """First inversion moves root up an octave."""
        chord = build_chord("c", "major", inversion=1)
        # First inversion: E G C
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["e", "g", "c"]
        # Check octaves
        octaves = [n.octave for n in chord.notes]
        assert octaves[2] > octaves[0]  # C is higher than E

    def test_unknown_chord_raises(self):
        """Unknown chord type should raise ValueError."""
        with pytest.raises(ValueError, match="Unknown chord type"):
            build_chord("c", "nonexistent")


class TestTriadConstructors:
    """Tests for triad constructor functions."""

    def test_major_function(self):
        chord = major("c")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["c", "e", "g"]

    def test_minor_function(self):
        chord = minor("a")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["a", "c", "e"]

    def test_dim_function(self):
        chord = dim("b")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["b", "d", "f"]

    def test_aug_function(self):
        chord = aug("c")
        pitches = [n.pitch for n in chord.notes]
        accidentals = [n.accidental for n in chord.notes]
        assert pitches == ["c", "e", "g"]
        assert accidentals == [None, None, "+"]  # G# is stored as g with + accidental

    def test_sus2_function(self):
        chord = sus2("c")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["c", "d", "g"]

    def test_sus4_function(self):
        chord = sus4("c")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["c", "f", "g"]


class TestSeventhChords:
    """Tests for seventh chord constructors."""

    def test_maj7(self):
        chord = maj7("c")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["c", "e", "g", "b"]

    def test_min7(self):
        chord = min7("a")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["a", "c", "e", "g"]

    def test_dom7(self):
        chord = dom7("g")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["g", "b", "d", "f"]

    def test_dim7(self):
        chord = dim7("b")
        # B D F Ab (Ab is g+ in our notation)
        assert len(chord.notes) == 4

    def test_half_dim7(self):
        chord = half_dim7("b")
        assert len(chord.notes) == 4

    def test_aug7(self):
        chord = aug7("c")
        assert len(chord.notes) == 4


class TestExtendedChords:
    """Tests for extended chord constructors."""

    def test_dom9(self):
        chord = dom9("g")
        assert len(chord.notes) == 5

    def test_maj9(self):
        chord = maj9("c")
        assert len(chord.notes) == 5

    def test_min9(self):
        chord = min9("a")
        assert len(chord.notes) == 5

    def test_add9(self):
        chord = add9("c")
        # Major triad plus 9th (no 7th)
        assert len(chord.notes) == 4

    def test_power(self):
        chord = power("e")
        pitches = [n.pitch for n in chord.notes]
        assert pitches == ["e", "b"]


class TestArpeggiate:
    """Tests for arpeggiate function."""

    def test_default_pattern(self):
        """Default pattern plays notes in order."""
        chord = major("c")
        arp = arpeggiate(chord)
        pitches = [n.pitch for n in arp]
        assert pitches == ["c", "e", "g"]

    def test_custom_pattern(self):
        """Custom pattern plays notes in specified order."""
        chord = major("c")
        arp = arpeggiate(chord, [0, 1, 2, 1])
        pitches = [n.pitch for n in arp]
        assert pitches == ["c", "e", "g", "e"]

    def test_with_duration(self):
        """Duration applied to all notes."""
        chord = major("c")
        arp = arpeggiate(chord, duration=16)
        for n in arp:
            assert n.duration == 16


class TestInvertChord:
    """Tests for invert_chord function."""

    def test_first_inversion(self):
        """First inversion moves root up."""
        chord = major("c")
        inverted = invert_chord(chord, 1)
        # Should have E G C (with C an octave higher)
        pitches = [n.pitch for n in inverted.notes]
        assert pitches == ["e", "g", "c"]

    def test_zero_inversion_returns_same(self):
        """Inversion 0 returns original chord."""
        chord = major("c")
        inverted = invert_chord(chord, 0)
        assert inverted == chord


class TestVoicing:
    """Tests for voicing function."""

    def test_spread_voicing(self):
        """Spread voicing puts notes in different octaves."""
        chord = major("c")
        voiced = voicing(chord, [3, 4, 5])
        octaves = [n.octave for n in voiced.notes]
        assert octaves == [3, 4, 5]

    def test_wrong_length_raises(self):
        """Wrong number of octaves should raise ValueError."""
        chord = major("c")
        with pytest.raises(ValueError):
            voicing(chord, [3, 4])  # Only 2 octaves for 3 notes


class TestListChordTypes:
    """Tests for list_chord_types function."""

    def test_returns_list(self):
        result = list_chord_types()
        assert isinstance(result, list)

    def test_contains_common_types(self):
        result = list_chord_types()
        assert "major" in result
        assert "minor" in result
        assert "dominant7" in result


# =============================================================================
# Cram Tests
# =============================================================================


class TestCram:
    """Tests for Cram class (tuplets)."""

    def test_cram_creation(self):
        """Basic Cram creation."""
        c = cram(note("c"), note("d"), note("e"), duration=4)
        assert len(c.elements) == 3
        assert c.duration == 4

    def test_cram_to_alda(self):
        """Cram should generate Alda syntax."""
        c = cram(note("c"), note("d"), note("e"), duration=4)
        result = c.to_alda()
        assert "{" in result
        assert "}" in result
        assert "4" in result

    def test_cram_class_directly(self):
        """Using Cram class directly."""
        c = Cram(elements=[note("c"), note("d")], duration=8)
        assert c.duration == 8

    def test_cram_to_ast(self):
        """Cram should produce valid AST node."""
        c = cram(note("c"), note("d"), duration=4)
        ast = c.to_ast()
        assert ast is not None


# =============================================================================
# Voice Tests
# =============================================================================


class TestVoice:
    """Tests for Voice class."""

    def test_voice_creation(self):
        """Basic Voice creation."""
        v = voice(1, note("c"), note("d"))
        assert v.number == 1
        assert len(v.elements) == 2

    def test_voice_to_alda(self):
        """Voice should generate Alda syntax."""
        v = voice(1, note("c"))
        result = v.to_alda()
        assert "V1:" in result

    def test_voice_zero_ends_voice(self):
        """Voice 0 ends the voice section."""
        v = voice(0)
        result = v.to_alda()
        assert "V0:" in result

    def test_voice_class_directly(self):
        """Using Voice class directly."""
        v = Voice(number=2, elements=[note("e"), note("f")])
        assert v.number == 2


class TestVoiceGroup:
    """Tests for VoiceGroup class."""

    def test_voice_group_creation(self):
        """VoiceGroup with multiple voices."""
        v1 = voice(1, note("c"))
        v2 = voice(2, note("e"))
        vg = voice_group(v1, v2)
        assert len(vg.voices) == 2

    def test_voice_group_to_alda(self):
        """VoiceGroup should generate Alda syntax."""
        v1 = voice(1, note("c"))
        v2 = voice(2, note("e"))
        vg = voice_group(v1, v2)
        result = vg.to_alda()
        assert "V1:" in result
        assert "V2:" in result


# =============================================================================
# Variable Tests
# =============================================================================


class TestVariable:
    """Tests for Variable class."""

    def test_variable_creation(self):
        """Basic Variable creation."""
        v = var("riff", note("c"), note("d"), note("e"))
        assert v.name == "riff"
        assert len(v.elements) == 3

    def test_variable_to_alda(self):
        """Variable should generate Alda syntax."""
        v = var("melody", note("c"))
        result = v.to_alda()
        assert "melody" in result
        assert "=" in result

    def test_variable_class_directly(self):
        """Using Variable class directly."""
        v = Variable(name="theme", elements=[note("g")])
        assert v.name == "theme"


class TestVariableRef:
    """Tests for VariableRef class."""

    def test_variable_ref_creation(self):
        """Basic VariableRef creation."""
        ref = var_ref("riff")
        assert ref.name == "riff"

    def test_variable_ref_to_alda(self):
        """VariableRef should generate Alda syntax."""
        ref = var_ref("melody")
        result = ref.to_alda()
        assert "melody" in result

    def test_variable_ref_class_directly(self):
        """Using VariableRef class directly."""
        ref = VariableRef(name="theme")
        assert ref.name == "theme"


# =============================================================================
# Marker Tests
# =============================================================================


class TestMarker:
    """Tests for Marker class."""

    def test_marker_creation(self):
        """Basic Marker creation."""
        m = marker("chorus")
        assert m.name == "chorus"

    def test_marker_to_alda(self):
        """Marker should generate Alda syntax."""
        m = marker("verse")
        result = m.to_alda()
        assert "%" in result
        assert "verse" in result

    def test_marker_class_directly(self):
        """Using Marker class directly."""
        m = Marker(name="bridge")
        assert m.name == "bridge"


class TestAtMarker:
    """Tests for AtMarker class."""

    def test_at_marker_creation(self):
        """Basic AtMarker creation."""
        am = at_marker("chorus")
        assert am.name == "chorus"

    def test_at_marker_to_alda(self):
        """AtMarker should generate Alda syntax."""
        am = at_marker("verse")
        result = am.to_alda()
        assert "@" in result
        assert "verse" in result

    def test_at_marker_class_directly(self):
        """Using AtMarker class directly."""
        am = AtMarker(name="bridge")
        assert am.name == "bridge"


# =============================================================================
# Integration Tests
# =============================================================================


class TestScaleChordIntegration:
    """Integration tests for scales and chords."""

    def test_scale_to_chord_progression(self):
        """Use scale to build chord progression."""
        pitches = scale("c", "major")
        # I-IV-V-I progression
        chords = [
            major(pitches[0]),  # C
            major(pitches[3]),  # F
            major(pitches[4]),  # G
            major(pitches[0]),  # C
        ]
        assert len(chords) == 4

    def test_arpeggiated_sequence(self):
        """Arpeggiate chords into a sequence."""
        chord = maj7("c")
        arp = arpeggiate(chord, [0, 1, 2, 3, 2, 1], duration=16)
        s = seq(*arp)
        assert len(s.elements) == 6

    def test_voice_with_scale(self):
        """Create voices from scales."""
        melody = voice(1, *scale_notes("c", "major", duration=4).elements)
        bass = voice(2, *scale_notes("c", "major", octave=3, duration=2).elements)
        vg = voice_group(melody, bass)
        assert len(vg.voices) == 2

    def test_variable_with_chord_sequence(self):
        """Store chord sequence in variable."""
        progression = var(
            "progression",
            major("c", duration=2),
            minor("a", duration=2),
            major("f", duration=2),
            major("g", duration=2),
        )
        assert progression.name == "progression"
        assert len(progression.elements) == 4



"""
Wcwidth module.

https://github.com/jquast/wcwidth
"""
# re-export all functions & definitions, even private ones, from top-level
# module path, to allow for 'from wcwidth import _private_func'.  Of course,
# user beware that any _private function may disappear or change signature at
# any future version.

# local
from .wcwidth import ZERO_WIDTH  # noqa
from .wcwidth import (VS16_NARROW_TO_WIDE, WIDE_EASTASIAN, _bisearch,
                      _wcmatch_version, _wcversion_value, list_versions,
                      wcswidth, wcwidth)

# The __all__ attribute defines the items exported from statement,
# 'from wcwidth import *', but also to say, "This is the public API".
__all__ = ("wcwidth", "wcswidth", "list_versions")

# We also used pkg_resources to load unicode version tables from version.json,
# generated by bin/update-tables.py, but some environments are unable to
# import pkg_resources for one reason or another, yikes!
__version__ = "0.2.14"



#!/usr/bin/env python3
#
# This script collects a list of symbols that are considered to be part of the
# CPython API. The result is used to inform the macOS linker that it's fine for
# those symbols to be undefined when an extension module is linked, as they
# will be provided when the extension module is loaded into the interpreter.

from urllib.request import urlopen
import re

funcs: "set[str]" = set()

for ver in ['3.9']:
  url = f'https://raw.githubusercontent.com/python/cpython/{ver}/PC/python3.def'
  output = urlopen(url).read().decode('utf-8')
  for match in re.findall(r"  (.*)=.*", output):
      funcs.add(match)

for ver in ['3.10', '3.11', 'main']:
    url = f'https://raw.githubusercontent.com/python/cpython/{ver}/PC/python3dll.c'
    output = urlopen(url).read().decode('utf-8')
    for match in re.findall(r"EXPORT_FUNC\((.*)\)", output):
        funcs.add(match)

funcs.remove('name')

# Add a few more functions that nanobind uses and which aren't in the above list
funcs |= {
    'PyFrame_GetBack',
    'PyGILState_Check',
    'PyObject_LengthHint',
    'Py_CompileStringExFlags',
    '_PyInterpreterState_Get',
    '_PyObject_MakeTpCall',
    '_PyObject_NextNotImplemented',
    '_Py_CheckFunctionResult',
    '_Py_RefTotal'
}

with open("darwin-ld-cpython.sym", "w") as f:
    for func in sorted(list(funcs)):
        f.write(f'-U _{func}\n')



"""CLI regression tests."""

import builtins

import pytest

from aldakit import __version__
from aldakit.cli import (
    create_parser,
    stdin_mode,
    _resolve_port_specifier,
    _resolve_output_port,
    _resolve_input_port,
)


def test_cli_version_matches_package(capsys):
    parser = create_parser()
    with pytest.raises(SystemExit):
        parser.parse_args(["--version"])

    out = capsys.readouterr().out.strip()
    assert out.endswith(__version__)


def test_stdin_mode_uses_backend_context(monkeypatch):
    entered = False
    exited = False

    class DummyBackend:
        def __init__(self, port_name=None):
            pass

        def __enter__(self):
            nonlocal entered
            entered = True
            return self

        def __exit__(self, exc_type, exc_val, exc_tb):
            nonlocal exited
            exited = True

        def play(self, sequence):
            return None

        def is_playing(self):
            return False

    def fake_input(prompt: str | None = None):
        raise EOFError

    monkeypatch.setattr("aldakit.cli.LibremidiBackend", DummyBackend)
    monkeypatch.setattr(builtins, "input", fake_input)

    assert stdin_mode(port_name=None, verbose=False) == 0
    assert entered and exited


class TestResolvePortSpecifier:
    """Tests for _resolve_port_specifier."""

    def test_none_returns_none_with_multiple_ports(self):
        port, ok = _resolve_port_specifier(None, ["PortA", "PortB"], "output")
        assert port is None
        assert ok is True

    def test_none_returns_none_with_no_ports(self):
        port, ok = _resolve_port_specifier(None, [], "output")
        assert port is None
        assert ok is True

    def test_index_resolves_to_name(self):
        port, ok = _resolve_port_specifier("0", ["FluidSynth", "IAC"], "output")
        assert port == "FluidSynth"
        assert ok is True

    def test_index_second_port(self):
        port, ok = _resolve_port_specifier("1", ["FluidSynth", "IAC"], "output")
        assert port == "IAC"
        assert ok is True

    def test_index_out_of_range_fails(self, capsys):
        port, ok = _resolve_port_specifier("5", ["A", "B"], "output")
        assert port is None
        assert ok is False
        err = capsys.readouterr().err
        assert "out of range" in err

    def test_name_passed_through(self):
        port, ok = _resolve_port_specifier("FluidSynth", ["FluidSynth"], "output")
        assert port == "FluidSynth"
        assert ok is True

    def test_partial_name_passed_through(self):
        # Backend handles partial matching, so we just pass it through
        port, ok = _resolve_port_specifier("Fluid", ["FluidSynth"], "output")
        assert port == "Fluid"
        assert ok is True


class TestResolveOutputPort:
    """Tests for _resolve_output_port auto-selection."""

    def test_auto_selects_single_port(self, monkeypatch):
        class DummyBackend:
            def list_output_ports(self):
                return ["OnlyPort"]

        monkeypatch.setattr("aldakit.cli.LibremidiBackend", DummyBackend)
        port, ok = _resolve_output_port(None)
        assert port == "OnlyPort"
        assert ok is True

    def test_no_auto_select_with_multiple_ports(self, monkeypatch):
        class DummyBackend:
            def list_output_ports(self):
                return ["PortA", "PortB"]

        monkeypatch.setattr("aldakit.cli.LibremidiBackend", DummyBackend)
        port, ok = _resolve_output_port(None)
        assert port is None
        assert ok is True

    def test_no_auto_select_with_no_ports(self, monkeypatch):
        class DummyBackend:
            def list_output_ports(self):
                return []

        monkeypatch.setattr("aldakit.cli.LibremidiBackend", DummyBackend)
        port, ok = _resolve_output_port(None)
        assert port is None
        assert ok is True


class TestResolveInputPort:
    """Tests for _resolve_input_port auto-selection."""

    def test_auto_selects_single_port(self, monkeypatch):
        monkeypatch.setattr(
            "aldakit.midi.transcriber.list_input_ports", lambda: ["OnlyInputPort"]
        )
        port, ok = _resolve_input_port(None)
        assert port == "OnlyInputPort"
        assert ok is True

    def test_no_auto_select_with_multiple_ports(self, monkeypatch):
        monkeypatch.setattr(
            "aldakit.midi.transcriber.list_input_ports", lambda: ["InputA", "InputB"]
        )
        port, ok = _resolve_input_port(None)
        assert port is None
        assert ok is True



"""libremidi-based MIDI backend for realtime playback."""

import threading
import time
from pathlib import Path

from ..types import MidiSequence
from .base import MidiBackend


class LibremidiBackend(MidiBackend):
    """MIDI backend using the libremidi library via nanobind.

    Provides low-latency realtime playback using the native libremidi library.
    """

    def __init__(self, port_name: str | None = None) -> None:
        """Initialize the libremidi backend.

        Args:
            port_name: Name of the MIDI output port to use.
                If None, will use a virtual port or first available port.
        """
        from ... import _libremidi

        self._libremidi = _libremidi
        self._port_name = port_name
        self._midi_out: _libremidi.MidiOut | None = None
        self._observer: _libremidi.Observer | None = None
        self._port_opened = False
        self._playing = False
        self._play_thread: threading.Thread | None = None
        self._stop_event = threading.Event()

    def _ensure_port_open(self) -> None:
        """Ensure the MIDI output port is open."""
        if self._port_opened and self._midi_out is not None:
            return

        self._midi_out = self._libremidi.MidiOut()
        self._observer = self._libremidi.Observer()
        ports = self._observer.get_output_ports()

        if self._port_name is not None:
            # Find port by name
            for port in ports:
                if (
                    port.port_name == self._port_name
                    or port.display_name == self._port_name
                ):
                    self._midi_out.open_port(port)
                    self._port_opened = True
                    return
            raise RuntimeError(
                f"Port '{self._port_name}' not found. "
                f"Available ports: {[p.display_name for p in ports]}"
            )
        elif ports:
            # Use first available port
            self._midi_out.open_port(ports[0])
            self._port_opened = True
        else:
            # Create a virtual port
            self._midi_out.open_virtual_port("AldakitMIDI")
            self._port_opened = True

    def _send_note_on(self, channel: int, note: int, velocity: int) -> None:
        """Send a note on message."""
        if self._midi_out is None:
            return
        status = 0x90 | (channel & 0x0F)
        self._midi_out.send_message(status, note & 0x7F, velocity & 0x7F)

    def _send_note_off(self, channel: int, note: int) -> None:
        """Send a note off message."""
        if self._midi_out is None:
            return
        status = 0x80 | (channel & 0x0F)
        self._midi_out.send_message(status, note & 0x7F, 0)

    def _send_program_change(self, channel: int, program: int) -> None:
        """Send a program change message."""
        if self._midi_out is None:
            return
        status = 0xC0 | (channel & 0x0F)
        self._midi_out.send_message(status, program & 0x7F)

    def _send_control_change(self, channel: int, control: int, value: int) -> None:
        """Send a control change message."""
        if self._midi_out is None:
            return
        status = 0xB0 | (channel & 0x0F)
        self._midi_out.send_message(status, control & 0x7F, value & 0x7F)

    def _send_all_notes_off(self) -> None:
        """Send all notes off on all channels."""
        for channel in range(16):
            self._send_control_change(channel, 123, 0)  # All Notes Off

    def play(self, sequence: MidiSequence) -> None:
        """Play a MIDI sequence in realtime.

        Args:
            sequence: The MIDI sequence to play.
        """
        self.stop()
        self._ensure_port_open()

        self._stop_event.clear()
        self._playing = True

        def play_thread():
            try:
                # Build a timeline of events
                events: list[tuple[float, str, tuple]] = []

                # Add program changes
                for pc in sequence.program_changes:
                    events.append((pc.time, "program", (pc.channel, pc.program)))

                # Add control changes
                for cc in sequence.control_changes:
                    events.append(
                        (cc.time, "control", (cc.channel, cc.control, cc.value))
                    )

                # Add note on/off events
                for note in sequence.notes:
                    events.append(
                        (
                            note.start_time,
                            "note_on",
                            (note.channel, note.pitch, note.velocity),
                        )
                    )
                    events.append(
                        (
                            note.start_time + note.duration,
                            "note_off",
                            (note.channel, note.pitch),
                        )
                    )

                # Sort by time, with note_off before note_on at same time
                events.sort(key=lambda e: (e[0], e[1] != "note_off"))

                # Play events
                start_time = time.perf_counter()
                for event_time, event_type, args in events:
                    if self._stop_event.is_set():
                        self._send_all_notes_off()
                        break

                    # Wait until event time
                    target_time = start_time + event_time
                    current_time = time.perf_counter()
                    if target_time > current_time:
                        # Use a loop with short sleeps to allow for stop events
                        while time.perf_counter() < target_time:
                            if self._stop_event.is_set():
                                self._send_all_notes_off()
                                return
                            remaining = target_time - time.perf_counter()
                            if remaining > 0.01:
                                time.sleep(0.01)
                            elif remaining > 0:
                                time.sleep(remaining)

                    # Send the event
                    if event_type == "note_on":
                        self._send_note_on(*args)
                    elif event_type == "note_off":
                        self._send_note_off(*args)
                    elif event_type == "program":
                        self._send_program_change(*args)
                    elif event_type == "control":
                        self._send_control_change(*args)

            finally:
                self._playing = False

        self._play_thread = threading.Thread(target=play_thread, daemon=True)
        self._play_thread.start()

    def stop(self) -> None:
        """Stop any currently playing sequence."""
        if self._playing:
            self._stop_event.set()
            if self._play_thread and self._play_thread.is_alive():
                self._play_thread.join(timeout=1.0)
            if self._port_opened:
                self._send_all_notes_off()
            self._playing = False

    def save(self, sequence: MidiSequence, path: Path | str) -> None:
        """Save a MIDI sequence to a Standard MIDI File.

        Args:
            sequence: The MIDI sequence to save.
            path: The output file path.
        """
        from ..smf import write_midi_file

        write_midi_file(sequence, path)

    def is_playing(self) -> bool:
        """Check if a sequence is currently playing."""
        return self._playing

    def close(self) -> None:
        """Close the MIDI output port."""
        self.stop()
        if self._midi_out is not None and self._port_opened:
            self._midi_out.close_port()
            self._port_opened = False

    def __enter__(self) -> "LibremidiBackend":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def list_output_ports(self) -> list[str]:
        """List available MIDI output ports.

        Returns:
            List of available MIDI output port names.
        """
        if self._observer is None:
            self._observer = self._libremidi.Observer()
        return [p.display_name for p in self._observer.get_output_ports()]



from __future__ import annotations

import xml.dom.minidom as minidom
from string import Formatter
from typing import Any

from .base import FormattedText, StyleAndTextTuples

__all__ = ["HTML"]


class HTML:
    """
    HTML formatted text.
    Take something HTML-like, for use as a formatted string.

    ::

        # Turn something into red.
        HTML('<style fg="ansired" bg="#00ff44">...</style>')

        # Italic, bold, underline and strike.
        HTML('<i>...</i>')
        HTML('<b>...</b>')
        HTML('<u>...</u>')
        HTML('<s>...</s>')

    All HTML elements become available as a "class" in the style sheet.
    E.g. ``<username>...</username>`` can be styled, by setting a style for
    ``username``.
    """

    def __init__(self, value: str) -> None:
        self.value = value
        document = minidom.parseString(f"<html-root>{value}</html-root>")

        result: StyleAndTextTuples = []
        name_stack: list[str] = []
        fg_stack: list[str] = []
        bg_stack: list[str] = []

        def get_current_style() -> str:
            "Build style string for current node."
            parts = []
            if name_stack:
                parts.append("class:" + ",".join(name_stack))

            if fg_stack:
                parts.append("fg:" + fg_stack[-1])
            if bg_stack:
                parts.append("bg:" + bg_stack[-1])
            return " ".join(parts)

        def process_node(node: Any) -> None:
            "Process node recursively."
            for child in node.childNodes:
                if child.nodeType == child.TEXT_NODE:
                    result.append((get_current_style(), child.data))
                else:
                    add_to_name_stack = child.nodeName not in (
                        "#document",
                        "html-root",
                        "style",
                    )
                    fg = bg = ""

                    for k, v in child.attributes.items():
                        if k == "fg":
                            fg = v
                        if k == "bg":
                            bg = v
                        if k == "color":
                            fg = v  # Alias for 'fg'.

                    # Check for spaces in attributes. This would result in
                    # invalid style strings otherwise.
                    if " " in fg:
                        raise ValueError('"fg" attribute contains a space.')
                    if " " in bg:
                        raise ValueError('"bg" attribute contains a space.')

                    if add_to_name_stack:
                        name_stack.append(child.nodeName)
                    if fg:
                        fg_stack.append(fg)
                    if bg:
                        bg_stack.append(bg)

                    process_node(child)

                    if add_to_name_stack:
                        name_stack.pop()
                    if fg:
                        fg_stack.pop()
                    if bg:
                        bg_stack.pop()

        process_node(document)

        self.formatted_text = FormattedText(result)

    def __repr__(self) -> str:
        return f"HTML({self.value!r})"

    def __pt_formatted_text__(self) -> StyleAndTextTuples:
        return self.formatted_text

    def format(self, *args: object, **kwargs: object) -> HTML:
        """
        Like `str.format`, but make sure that the arguments are properly
        escaped.
        """
        return HTML(FORMATTER.vformat(self.value, args, kwargs))

    def __mod__(self, value: object) -> HTML:
        """
        HTML('<b>%s</b>') % value
        """
        if not isinstance(value, tuple):
            value = (value,)

        value = tuple(html_escape(i) for i in value)
        return HTML(self.value % value)


class HTMLFormatter(Formatter):
    def format_field(self, value: object, format_spec: str) -> str:
        return html_escape(format(value, format_spec))


def html_escape(text: object) -> str:
    # The string interpolation functions also take integers and other types.
    # Convert to string first.
    if not isinstance(text, str):
        text = f"{text}"

    return (
        text.replace("&", "&amp;")
        .replace("<", "&lt;")
        .replace(">", "&gt;")
        .replace('"', "&quot;")
    )


FORMATTER = HTMLFormatter()



"""
Utility for running a prompt_toolkit application in an asyncssh server.
"""

from __future__ import annotations

import asyncio
import traceback
from asyncio import get_running_loop
from typing import Any, Callable, Coroutine, TextIO, cast

import asyncssh
from prompt_toolkit.application.current import AppSession, create_app_session
from prompt_toolkit.data_structures import Size
from prompt_toolkit.input import PipeInput, create_pipe_input
from prompt_toolkit.output.vt100 import Vt100_Output

__all__ = ["PromptToolkitSSHSession", "PromptToolkitSSHServer"]


class PromptToolkitSSHSession(asyncssh.SSHServerSession):  # type: ignore
    def __init__(
        self,
        interact: Callable[[PromptToolkitSSHSession], Coroutine[Any, Any, None]],
        *,
        enable_cpr: bool,
    ) -> None:
        self.interact = interact
        self.enable_cpr = enable_cpr
        self.interact_task: asyncio.Task[None] | None = None
        self._chan: Any | None = None
        self.app_session: AppSession | None = None

        # PipInput object, for sending input in the CLI.
        # (This is something that we can use in the prompt_toolkit event loop,
        # but still write date in manually.)
        self._input: PipeInput | None = None
        self._output: Vt100_Output | None = None

        # Output object. Don't render to the real stdout, but write everything
        # in the SSH channel.
        class Stdout:
            def write(s, data: str) -> None:
                try:
                    if self._chan is not None:
                        self._chan.write(data.replace("\n", "\r\n"))
                except BrokenPipeError:
                    pass  # Channel not open for sending.

            def isatty(s) -> bool:
                return True

            def flush(s) -> None:
                pass

            @property
            def encoding(s) -> str:
                assert self._chan is not None
                return str(self._chan._orig_chan.get_encoding()[0])

        self.stdout = cast(TextIO, Stdout())

    def _get_size(self) -> Size:
        """
        Callable that returns the current `Size`, required by Vt100_Output.
        """
        if self._chan is None:
            return Size(rows=20, columns=79)
        else:
            width, height, pixwidth, pixheight = self._chan.get_terminal_size()
            return Size(rows=height, columns=width)

    def connection_made(self, chan: Any) -> None:
        self._chan = chan

    def shell_requested(self) -> bool:
        return True

    def session_started(self) -> None:
        self.interact_task = get_running_loop().create_task(self._interact())

    async def _interact(self) -> None:
        if self._chan is None:
            # Should not happen.
            raise Exception("`_interact` called before `connection_made`.")

        if hasattr(self._chan, "set_line_mode") and self._chan._editor is not None:
            # Disable the line editing provided by asyncssh. Prompt_toolkit
            # provides the line editing.
            self._chan.set_line_mode(False)

        term = self._chan.get_terminal_type()

        self._output = Vt100_Output(
            self.stdout, self._get_size, term=term, enable_cpr=self.enable_cpr
        )

        with create_pipe_input() as self._input:
            with create_app_session(input=self._input, output=self._output) as session:
                self.app_session = session
                try:
                    await self.interact(self)
                except BaseException:
                    traceback.print_exc()
                finally:
                    # Close the connection.
                    self._chan.close()
                    self._input.close()

    def terminal_size_changed(
        self, width: int, height: int, pixwidth: object, pixheight: object
    ) -> None:
        # Send resize event to the current application.
        if self.app_session and self.app_session.app:
            self.app_session.app._on_resize()

    def data_received(self, data: str, datatype: object) -> None:
        if self._input is None:
            # Should not happen.
            return

        self._input.send_text(data)


class PromptToolkitSSHServer(asyncssh.SSHServer):
    """
    Run a prompt_toolkit application over an asyncssh server.

    This takes one argument, an `interact` function, which is called for each
    connection. This should be an asynchronous function that runs the
    prompt_toolkit applications. This function runs in an `AppSession`, which
    means that we can have multiple UI interactions concurrently.

    Example usage:

    .. code:: python

        async def interact(ssh_session: PromptToolkitSSHSession) -> None:
            await yes_no_dialog("my title", "my text").run_async()

            prompt_session = PromptSession()
            text = await prompt_session.prompt_async("Type something: ")
            print_formatted_text('You said: ', text)

        server = PromptToolkitSSHServer(interact=interact)
        loop = get_running_loop()
        loop.run_until_complete(
            asyncssh.create_server(
                lambda: MySSHServer(interact),
                "",
                port,
                server_host_keys=["/etc/ssh/..."],
            )
        )
        loop.run_forever()

    :param enable_cpr: When `True`, the default, try to detect whether the SSH
        client runs in a terminal that responds to "cursor position requests".
        That way, we can properly determine how much space there is available
        for the UI (especially for drop down menus) to render.
    """

    def __init__(
        self,
        interact: Callable[[PromptToolkitSSHSession], Coroutine[Any, Any, None]],
        *,
        enable_cpr: bool = True,
    ) -> None:
        self.interact = interact
        self.enable_cpr = enable_cpr

    def begin_auth(self, username: str) -> bool:
        # No authentication.
        return False

    def session_requested(self) -> PromptToolkitSSHSession:
        return PromptToolkitSSHSession(self.interact, enable_cpr=self.enable_cpr)



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, List, TypeVar, cast, overload

from prompt_toolkit.formatted_text.base import OneStyleAndTextTuple

if TYPE_CHECKING:
    from typing_extensions import SupportsIndex

__all__ = [
    "explode_text_fragments",
]

_T = TypeVar("_T", bound=OneStyleAndTextTuple)


class _ExplodedList(List[_T]):
    """
    Wrapper around a list, that marks it as 'exploded'.

    As soon as items are added or the list is extended, the new items are
    automatically exploded as well.
    """

    exploded = True

    def append(self, item: _T) -> None:
        self.extend([item])

    def extend(self, lst: Iterable[_T]) -> None:
        super().extend(explode_text_fragments(lst))

    def insert(self, index: SupportsIndex, item: _T) -> None:
        raise NotImplementedError  # TODO

    # TODO: When creating a copy() or [:], return also an _ExplodedList.

    @overload
    def __setitem__(self, index: SupportsIndex, value: _T) -> None: ...

    @overload
    def __setitem__(self, index: slice, value: Iterable[_T]) -> None: ...

    def __setitem__(
        self, index: SupportsIndex | slice, value: _T | Iterable[_T]
    ) -> None:
        """
        Ensure that when `(style_str, 'long string')` is set, the string will be
        exploded.
        """
        if not isinstance(index, slice):
            int_index = index.__index__()
            index = slice(int_index, int_index + 1)
        if isinstance(value, tuple):  # In case of `OneStyleAndTextTuple`.
            value = cast("List[_T]", [value])

        super().__setitem__(index, explode_text_fragments(value))


def explode_text_fragments(fragments: Iterable[_T]) -> _ExplodedList[_T]:
    """
    Turn a list of (style_str, text) tuples into another list where each string is
    exactly one character.

    It should be fine to call this function several times. Calling this on a
    list that is already exploded, is a null operation.

    :param fragments: List of (style, text) tuples.
    """
    # When the fragments is already exploded, don't explode again.
    if isinstance(fragments, _ExplodedList):
        return fragments

    result: list[_T] = []

    for style, string, *rest in fragments:
        for c in string:
            result.append((style, c, *rest))  # type: ignore

    return _ExplodedList(result)



"""Core compose elements: Note, Rest, Chord, Seq."""

from __future__ import annotations

from dataclasses import dataclass, field, replace
from typing import TYPE_CHECKING, Any

from .base import ComposeElement

if TYPE_CHECKING:
    pass

from ..ast_nodes import (
    AtMarkerNode,
    ChordNode,
    CramNode,
    DurationNode,
    EventSequenceNode,
    MarkerNode,
    NoteLengthMsNode,
    NoteLengthNode,
    NoteLengthSecondsNode,
    NoteNode,
    RepeatNode,
    RestNode,
    VariableDefinitionNode,
    VariableReferenceNode,
    VoiceGroupNode,
    VoiceNode,
)

# Pitch to semitone offset from C
_PITCH_OFFSETS = {"c": 0, "d": 2, "e": 4, "f": 5, "g": 7, "a": 9, "b": 11}
_SEMITONE_TO_PITCH = ["c", "c", "d", "d", "e", "f", "f", "g", "g", "a", "a", "b"]
_SEMITONE_ACCIDENTALS = ["", "+", "", "+", "", "", "+", "", "+", "", "+", ""]


@dataclass(frozen=True)
class Note(ComposeElement):
    """A musical note.

    All parameters except pitch are keyword-only to avoid ambiguity.

    Examples:
        >>> note("c")                              # C quarter note
        >>> note("c", duration=4)                  # C quarter note (explicit)
        >>> note("c", duration=8, accidental="+")  # C# eighth note
        >>> note("c", octave=5)                    # C in octave 5
        >>> note("c", ms=500)                      # C for 500 milliseconds
    """

    pitch: str
    duration: int | None = None
    octave: int | None = None
    accidental: str | None = None
    dots: int = 0
    ms: float | None = None
    seconds: float | None = None
    slurred: bool = False

    def __post_init__(self) -> None:
        # Validate pitch
        if self.pitch.lower() not in _PITCH_OFFSETS:
            raise ValueError(f"Invalid pitch: {self.pitch}. Must be a-g.")

    def to_ast(self) -> NoteNode:
        """Convert to AST NoteNode."""
        # Build accidentals list
        accidentals: list[str] = []
        if self.accidental:
            accidentals = list(self.accidental)

        # Build duration node
        duration_node = self._build_duration_node()

        return NoteNode(
            letter=self.pitch.lower(),
            accidentals=accidentals,
            duration=duration_node,
            slurred=self.slurred,
            position=None,
        )

    def _build_duration_node(self) -> DurationNode | None:
        """Build duration node from duration parameters."""
        if self.ms is not None:
            return DurationNode(
                components=[NoteLengthMsNode(ms=self.ms, position=None)],
                position=None,
            )
        elif self.seconds is not None:
            return DurationNode(
                components=[NoteLengthSecondsNode(seconds=self.seconds, position=None)],
                position=None,
            )
        elif self.duration is not None:
            return DurationNode(
                components=[
                    NoteLengthNode(
                        denominator=self.duration, dots=self.dots, position=None
                    )
                ],
                position=None,
            )
        elif self.dots > 0:
            # Dots without explicit duration - use default quarter note
            return DurationNode(
                components=[
                    NoteLengthNode(denominator=4, dots=self.dots, position=None)
                ],
                position=None,
            )
        return None

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        result = self.pitch.lower()

        # Accidentals
        if self.accidental:
            result += self.accidental

        # Duration
        if self.ms is not None:
            result += f"{int(self.ms)}ms"
        elif self.seconds is not None:
            result += f"{self.seconds}s"
        elif self.duration is not None:
            result += str(self.duration)
            result += "." * self.dots

        # Slur
        if self.slurred:
            result += "~"

        return result

    @property
    def midi_pitch(self) -> int:
        """Calculate MIDI pitch number.

        Uses octave 4 as default if not specified.
        """
        oct = self.octave if self.octave is not None else 4
        base = _PITCH_OFFSETS[self.pitch.lower()]

        # Apply accidentals
        offset = 0
        if self.accidental:
            for acc in self.accidental:
                if acc == "+":
                    offset += 1
                elif acc == "-":
                    offset -= 1
                # "_" (natural) doesn't change offset

        return (oct + 1) * 12 + base + offset

    # Transformation methods (return new Note instances)

    def sharpen(self) -> Note:
        """Return a new note raised by one semitone."""
        new_acc = (self.accidental or "") + "+"
        return replace(self, accidental=new_acc)

    def flatten(self) -> Note:
        """Return a new note lowered by one semitone."""
        new_acc = (self.accidental or "") + "-"
        return replace(self, accidental=new_acc)

    def transpose(self, semitones: int) -> Note:
        """Return a new note transposed by the given number of semitones."""
        current_midi = self.midi_pitch
        new_midi = current_midi + semitones

        # Calculate new octave and pitch
        new_octave = (new_midi // 12) - 1
        semitone_in_octave = new_midi % 12

        new_pitch = _SEMITONE_TO_PITCH[semitone_in_octave]
        new_accidental = _SEMITONE_ACCIDENTALS[semitone_in_octave] or None

        return replace(
            self, pitch=new_pitch, octave=new_octave, accidental=new_accidental
        )

    def with_duration(self, duration: int) -> Note:
        """Return a new note with the given duration."""
        return replace(self, duration=duration, ms=None, seconds=None)

    def with_octave(self, octave: int) -> Note:
        """Return a new note in the given octave."""
        return replace(self, octave=octave)

    def with_dots(self, dots: int) -> Note:
        """Return a new note with the given number of dots."""
        return replace(self, dots=dots)

    def slur(self) -> Note:
        """Return a new slurred note."""
        return replace(self, slurred=True)

    def __mul__(self, n: int) -> Repeat:
        """Repeat this note n times."""
        return Repeat(element=self, times=n)

    def __rmul__(self, n: int) -> Repeat:
        """Repeat this note n times (reverse order)."""
        return self.__mul__(n)


@dataclass(frozen=True)
class Rest(ComposeElement):
    """A musical rest (silence).

    Examples:
        >>> rest()                    # Quarter rest
        >>> rest(duration=2)          # Half rest
        >>> rest(ms=1000)             # One second rest
    """

    duration: int | None = None
    dots: int = 0
    ms: float | None = None
    seconds: float | None = None

    def to_ast(self) -> RestNode:
        """Convert to AST RestNode."""
        duration_node = self._build_duration_node()
        return RestNode(duration=duration_node, position=None)

    def _build_duration_node(self) -> DurationNode | None:
        """Build duration node from duration parameters."""
        if self.ms is not None:
            return DurationNode(
                components=[NoteLengthMsNode(ms=self.ms, position=None)],
                position=None,
            )
        elif self.seconds is not None:
            return DurationNode(
                components=[NoteLengthSecondsNode(seconds=self.seconds, position=None)],
                position=None,
            )
        elif self.duration is not None:
            return DurationNode(
                components=[
                    NoteLengthNode(
                        denominator=self.duration, dots=self.dots, position=None
                    )
                ],
                position=None,
            )
        return None

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        result = "r"

        if self.ms is not None:
            result += f"{int(self.ms)}ms"
        elif self.seconds is not None:
            result += f"{self.seconds}s"
        elif self.duration is not None:
            result += str(self.duration)
            result += "." * self.dots

        return result


@dataclass(frozen=True)
class Chord(ComposeElement):
    """A chord (multiple notes played simultaneously).

    Examples:
        >>> chord("c", "e", "g")                    # C major
        >>> chord(note("c"), note("e"), note("g"))  # Same, using Note objects
        >>> chord("c", "e", "g", duration=1)        # Whole note C major
    """

    notes: tuple[Note, ...] = field(default_factory=tuple)
    duration: int | None = None
    dots: int = 0

    def to_ast(self) -> ChordNode:
        """Convert to AST ChordNode."""
        # Apply chord duration to first note if specified
        note_asts = []
        for i, n in enumerate(self.notes):
            if i == 0 and self.duration is not None:
                modified_note = n.with_duration(self.duration)
                if self.dots:
                    modified_note = modified_note.with_dots(self.dots)
                note_asts.append(modified_note.to_ast())
            else:
                note_asts.append(n.to_ast())

        return ChordNode(notes=note_asts, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        parts = []
        for i, n in enumerate(self.notes):
            if i == 0 and self.duration is not None:
                modified_note = n.with_duration(self.duration)
                if self.dots:
                    modified_note = modified_note.with_dots(self.dots)
                parts.append(modified_note.to_alda())
            else:
                # Subsequent notes don't repeat duration
                parts.append(n.pitch + (n.accidental or ""))
        return "/".join(parts)


@dataclass
class Seq(ComposeElement):
    """A sequence of musical elements.

    Examples:
        >>> seq(note("c"), note("d"), note("e"))
        >>> seq.from_alda("c d e f g")
        >>> seq(note("c"), note("d")) * 4  # Repeat 4 times
    """

    elements: list[ComposeElement] = field(default_factory=list)
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_ast(self) -> EventSequenceNode:
        """Convert to AST EventSequenceNode."""
        events = [e.to_ast() for e in self.elements]
        return EventSequenceNode(events=events, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        return " ".join(e.to_alda() for e in self.elements)

    @classmethod
    def from_alda(cls, source: str) -> Seq:
        """Create a Seq by parsing Alda source code.

        Args:
            source: Alda source code snippet.

        Returns:
            Seq containing parsed elements.
        """
        from ..parser import parse

        ast = parse(source)
        # Wrap the AST in a ParsedSeq that delegates to_ast to the parsed result
        return _ParsedSeq(ast=ast, source=source)

    def __mul__(self, n: int) -> Repeat:
        """Repeat this sequence n times."""
        return Repeat(element=self, times=n)

    def __rmul__(self, n: int) -> Repeat:
        """Repeat this sequence n times (reverse order)."""
        return self.__mul__(n)

    def __add__(self, other: Seq) -> Seq:
        """Concatenate two sequences."""
        if isinstance(other, Seq):
            merged = dict(self.metadata)
            for key, value in other.metadata.items():
                merged.setdefault(key, value)
            return Seq(
                elements=list(self.elements) + list(other.elements), metadata=merged
            )
        return NotImplemented


@dataclass
class _ParsedSeq(Seq):
    """A sequence created from parsed Alda source.

    This wraps a parsed AST to avoid re-parsing.
    """

    ast: object = None  # RootNode
    source: str = ""
    elements: list[ComposeElement] = field(default_factory=list)
    metadata: dict[str, Any] = field(default_factory=dict)

    def to_ast(self) -> EventSequenceNode:
        """Return the parsed AST."""
        # The parsed AST is a RootNode, return its contents as EventSequenceNode
        from ..ast_nodes import EventSequenceNode as ESN
        from ..ast_nodes import RootNode

        if isinstance(self.ast, RootNode):
            # Flatten RootNode children into EventSequenceNode
            return ESN(events=list(self.ast.children), position=None)
        return ESN(events=[], position=None)

    def to_alda(self) -> str:
        """Return the original source."""
        return self.source


@dataclass(frozen=True)
class Repeat(ComposeElement):
    """A repeated element.

    Examples:
        >>> note("c") * 4        # Repeat note 4 times
        >>> seq(...) * 8         # Repeat sequence 8 times
    """

    element: ComposeElement
    times: int

    def to_ast(self) -> RepeatNode:
        """Convert to AST RepeatNode."""
        from ..ast_nodes import BracketedSequenceNode

        inner_ast = self.element.to_ast()

        # Wrap in bracketed sequence if it's an EventSequenceNode
        if isinstance(inner_ast, EventSequenceNode):
            bracketed = BracketedSequenceNode(events=inner_ast, position=None)
            return RepeatNode(event=bracketed, times=self.times, position=None)
        else:
            return RepeatNode(event=inner_ast, times=self.times, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        inner = self.element.to_alda()
        # Wrap in brackets if it's a sequence
        if isinstance(self.element, Seq) and len(self.element.elements) > 1:
            return f"[{inner}]*{self.times}"
        return f"{inner}*{self.times}"


# Factory functions


def note(
    pitch: str,
    *,
    duration: int | None = None,
    octave: int | None = None,
    accidental: str | None = None,
    dots: int = 0,
    ms: float | None = None,
    seconds: float | None = None,
    slurred: bool = False,
) -> Note:
    """Create a note.

    Args:
        pitch: Note pitch (a-g).
        duration: Note duration denominator (1=whole, 2=half, 4=quarter, etc.).
        octave: Octave number (typically 0-8).
        accidental: Accidental string ("+", "-", "_", "++", etc.).
        dots: Number of dots (0-3).
        ms: Duration in milliseconds.
        seconds: Duration in seconds.
        slurred: Whether the note is slurred to the next.

    Returns:
        Note element.
    """
    return Note(
        pitch=pitch,
        duration=duration,
        octave=octave,
        accidental=accidental,
        dots=dots,
        ms=ms,
        seconds=seconds,
        slurred=slurred,
    )


def rest(
    *,
    duration: int | None = None,
    dots: int = 0,
    ms: float | None = None,
    seconds: float | None = None,
) -> Rest:
    """Create a rest.

    Args:
        duration: Rest duration denominator (1=whole, 2=half, 4=quarter, etc.).
        dots: Number of dots (0-3).
        ms: Duration in milliseconds.
        seconds: Duration in seconds.

    Returns:
        Rest element.
    """
    return Rest(duration=duration, dots=dots, ms=ms, seconds=seconds)


def chord(
    *notes_or_pitches: Note | str, duration: int | None = None, dots: int = 0
) -> Chord:
    """Create a chord.

    Args:
        *notes_or_pitches: Note objects or pitch strings.
        duration: Chord duration (applied to first note).
        dots: Number of dots to apply to the chord duration.

    Returns:
        Chord element.
    """
    notes = []
    for item in notes_or_pitches:
        if isinstance(item, Note):
            notes.append(item)
        elif isinstance(item, str):
            notes.append(Note(pitch=item))
        else:
            raise TypeError(f"Expected Note or str, got {type(item)}")

    return Chord(notes=tuple(notes), duration=duration, dots=dots)


def seq(*elements: ComposeElement, metadata: dict[str, Any] | None = None) -> Seq:
    """Create a sequence.

    Args:
        *elements: Compose elements to include in the sequence.
        metadata: Optional metadata dictionary for timing/annotations.

    Returns:
        Seq element.
    """
    return Seq(elements=list(elements), metadata=dict(metadata or {}))


# =============================================================================
# Cram (Tuplets)
# =============================================================================


@dataclass(frozen=True)
class Cram(ComposeElement):
    """A cram expression (tuplet) - fit multiple notes into a duration.

    Cram expressions fit a group of notes into a specific duration,
    creating tuplets like triplets, quintuplets, etc.

    Examples:
        >>> cram(note("c"), note("d"), note("e"), duration=4)  # Triplet in quarter note
        >>> cram(note("c"), note("d"), note("e"), note("f"), note("g"), duration=2)  # Quintuplet
    """

    elements: list[ComposeElement]
    duration: int | None = None
    dots: int = 0

    def to_ast(self) -> CramNode:
        """Convert to AST CramNode."""
        # Build event sequence from elements
        events = []
        for elem in self.elements:
            events.append(elem.to_ast())

        event_seq = EventSequenceNode(events=events, position=None)

        # Build duration node
        duration_node = None
        if self.duration is not None:
            duration_node = DurationNode(
                components=[
                    NoteLengthNode(
                        denominator=self.duration, dots=self.dots, position=None
                    )
                ],
                position=None,
            )

        return CramNode(events=event_seq, duration=duration_node, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        inner = " ".join(elem.to_alda() for elem in self.elements)
        if self.duration is not None:
            dots = "." * self.dots
            return f"{{{inner}}}{self.duration}{dots}"
        return f"{{{inner}}}"


def cram(*elements: ComposeElement, duration: int | None = None, dots: int = 0) -> Cram:
    """Create a cram expression (tuplet).

    Args:
        *elements: Notes/rests to fit into the duration.
        duration: Duration denominator to fit notes into.
        dots: Number of dots on the duration.

    Returns:
        Cram element.

    Examples:
        >>> cram(note("c"), note("d"), note("e"), duration=4)  # Triplet
        >>> cram(note("c"), note("d"), note("e"), note("f"), note("g"), duration=2)
    """
    return Cram(elements=list(elements), duration=duration, dots=dots)


# =============================================================================
# Voice (Parallel Lines)
# =============================================================================


@dataclass(frozen=True)
class Voice(ComposeElement):
    """A voice within a part for polyphonic writing.

    Voices allow multiple parallel lines within a single part.
    Voice numbers start at 1; V0 ends the voice section.

    Examples:
        >>> voice(1, note("c"), note("d"), note("e"))  # Voice 1
        >>> voice(2, note("e"), note("f"), note("g"))  # Voice 2
    """

    number: int
    elements: list[ComposeElement]

    def __post_init__(self) -> None:
        if self.number < 0:
            raise ValueError("Voice number must be non-negative (0 ends voices)")

    def to_ast(self) -> VoiceNode:
        """Convert to AST VoiceNode."""
        events = [elem.to_ast() for elem in self.elements]
        event_seq = EventSequenceNode(events=events, position=None)
        return VoiceNode(number=self.number, events=event_seq, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        inner = " ".join(elem.to_alda() for elem in self.elements)
        return f"V{self.number}: {inner}"


@dataclass(frozen=True)
class VoiceGroup(ComposeElement):
    """A group of parallel voices.

    Examples:
        >>> voice_group(
        ...     voice(1, note("c"), note("d")),
        ...     voice(2, note("e"), note("f")),
        ... )
    """

    voices: list[Voice]

    def to_ast(self) -> VoiceGroupNode:
        """Convert to AST VoiceGroupNode."""
        voice_nodes = [v.to_ast() for v in self.voices]
        return VoiceGroupNode(voices=voice_nodes, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        lines = [v.to_alda() for v in self.voices]
        lines.append("V0:")  # End voices
        return "\n".join(lines)


def voice(number: int, *elements: ComposeElement) -> Voice:
    """Create a voice.

    Args:
        number: Voice number (1, 2, etc.; 0 ends voice section).
        *elements: Notes/rests in this voice.

    Returns:
        Voice element.
    """
    return Voice(number=number, elements=list(elements))


def voice_group(*voices: Voice) -> VoiceGroup:
    """Create a voice group from multiple voices.

    Args:
        *voices: Voice elements to group.

    Returns:
        VoiceGroup element.
    """
    return VoiceGroup(voices=list(voices))


# =============================================================================
# Variables
# =============================================================================


@dataclass(frozen=True)
class Variable(ComposeElement):
    """A variable definition.

    Variables allow naming a sequence for reuse.

    Examples:
        >>> var("riff", note("c"), note("d"), note("e"))
        >>> var_ref("riff")  # Use the variable
    """

    name: str
    elements: list[ComposeElement]

    def to_ast(self) -> VariableDefinitionNode:
        """Convert to AST VariableDefinitionNode."""
        events = [elem.to_ast() for elem in self.elements]
        event_seq = EventSequenceNode(events=events, position=None)
        return VariableDefinitionNode(name=self.name, events=event_seq, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        inner = " ".join(elem.to_alda() for elem in self.elements)
        return f"{self.name} = {inner}"


@dataclass(frozen=True)
class VariableRef(ComposeElement):
    """A reference to a previously defined variable.

    Examples:
        >>> var_ref("riff")  # Use a variable named 'riff'
    """

    name: str

    def to_ast(self) -> VariableReferenceNode:
        """Convert to AST VariableReferenceNode."""
        return VariableReferenceNode(name=self.name, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        return self.name


def var(name: str, *elements: ComposeElement) -> Variable:
    """Define a variable.

    Args:
        name: Variable name.
        *elements: Notes/rests to assign to the variable.

    Returns:
        Variable definition element.

    Examples:
        >>> var("riff", note("c"), note("d"), note("e"), note("f"))
    """
    return Variable(name=name, elements=list(elements))


def var_ref(name: str) -> VariableRef:
    """Reference a variable.

    Args:
        name: Variable name to reference.

    Returns:
        Variable reference element.

    Examples:
        >>> var_ref("riff")
    """
    return VariableRef(name=name)


# =============================================================================
# Markers
# =============================================================================


@dataclass(frozen=True)
class Marker(ComposeElement):
    """A marker definition for synchronization points.

    Markers allow different parts to synchronize at named points.

    Examples:
        >>> marker("chorus")  # Define marker
        >>> at_marker("chorus")  # Jump to marker in another part
    """

    name: str

    def to_ast(self) -> MarkerNode:
        """Convert to AST MarkerNode."""
        return MarkerNode(name=self.name, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        return f"%{self.name}"


@dataclass(frozen=True)
class AtMarker(ComposeElement):
    """A marker reference (jump to marker).

    Examples:
        >>> at_marker("chorus")  # Jump to the 'chorus' marker
    """

    name: str

    def to_ast(self) -> AtMarkerNode:
        """Convert to AST AtMarkerNode."""
        return AtMarkerNode(name=self.name, position=None)

    def to_alda(self) -> str:
        """Convert to Alda source code."""
        return f"@{self.name}"


def marker(name: str) -> Marker:
    """Create a marker.

    Args:
        name: Marker name.

    Returns:
        Marker element.

    Examples:
        >>> marker("verse")
        >>> marker("chorus")
    """
    return Marker(name=name)


def at_marker(name: str) -> AtMarker:
    """Create a marker reference (jump to marker).

    Args:
        name: Marker name to jump to.

    Returns:
        AtMarker element.

    Examples:
        >>> at_marker("chorus")  # Jump to chorus marker
    """
    return AtMarker(name=name)



"""Tests for the aldakit scanner."""

import pytest
from aldakit.scanner import Scanner
from aldakit.tokens import TokenType
from aldakit.errors import AldaScanError


class TestBasicTokens:
    """Test basic token recognition."""

    def test_note_letters(self):
        scanner = Scanner("a b c d e f g")
        tokens = scanner.scan()
        note_tokens = [t for t in tokens if t.type == TokenType.NOTE_LETTER]
        assert len(note_tokens) == 7
        assert [t.literal for t in note_tokens] == ["a", "b", "c", "d", "e", "f", "g"]

    def test_rest_letter(self):
        scanner = Scanner("r")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.REST_LETTER
        assert tokens[0].lexeme == "r"

    def test_accidentals(self):
        scanner = Scanner("+ - _")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert types == [TokenType.SHARP, TokenType.FLAT, TokenType.NATURAL]

    def test_octave_controls(self):
        scanner = Scanner("> < o4 o0 o9")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert types == [
            TokenType.OCTAVE_UP,
            TokenType.OCTAVE_DOWN,
            TokenType.OCTAVE_SET,
            TokenType.OCTAVE_SET,
            TokenType.OCTAVE_SET,
        ]
        # Check octave values
        octave_tokens = [t for t in tokens if t.type == TokenType.OCTAVE_SET]
        assert [t.literal for t in octave_tokens] == [4, 0, 9]


class TestDurations:
    """Test duration token recognition."""

    def test_simple_durations(self):
        scanner = Scanner("4 8 16 1 2")
        tokens = scanner.scan()
        duration_tokens = [t for t in tokens if t.type == TokenType.NOTE_LENGTH]
        assert len(duration_tokens) == 5
        assert [t.literal for t in duration_tokens] == [4, 8, 16, 1, 2]

    def test_millisecond_duration(self):
        scanner = Scanner("500ms 1000ms")
        tokens = scanner.scan()
        ms_tokens = [t for t in tokens if t.type == TokenType.NOTE_LENGTH_MS]
        assert len(ms_tokens) == 2
        assert [t.literal for t in ms_tokens] == [500.0, 1000.0]

    def test_second_duration(self):
        scanner = Scanner("2s 0.5s")
        tokens = scanner.scan()
        s_tokens = [t for t in tokens if t.type == TokenType.NOTE_LENGTH_SECONDS]
        assert len(s_tokens) == 2
        assert [t.literal for t in s_tokens] == [2.0, 0.5]

    def test_dotted_note(self):
        scanner = Scanner("4.")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.NOTE_LENGTH
        assert tokens[0].literal == 4
        assert tokens[1].type == TokenType.DOT


class TestStructuralTokens:
    """Test structural tokens."""

    def test_barline(self):
        scanner = Scanner("|")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.BARLINE

    def test_tie(self):
        scanner = Scanner("~")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.TIE

    def test_separator(self):
        scanner = Scanner("/")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.SEPARATOR

    def test_colon(self):
        scanner = Scanner(":")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.COLON


class TestPartDeclaration:
    """Test part declaration tokens."""

    def test_simple_part(self):
        scanner = Scanner("piano:")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.NAME
        assert tokens[0].literal == "piano"
        assert tokens[1].type == TokenType.COLON

    def test_part_with_alias(self):
        scanner = Scanner('violin "v1":')
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.NAME
        assert tokens[0].literal == "violin"
        assert tokens[1].type == TokenType.ALIAS
        assert tokens[1].literal == "v1"
        assert tokens[2].type == TokenType.COLON

    def test_multi_instrument(self):
        scanner = Scanner("violin/viola:")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert types == [
            TokenType.NAME,
            TokenType.SEPARATOR,
            TokenType.NAME,
            TokenType.COLON,
        ]


class TestSExpressions:
    """Test S-expression (Lisp) tokenization."""

    def test_simple_sexp(self):
        scanner = Scanner("(tempo 120)")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert types == [
            TokenType.LEFT_PAREN,
            TokenType.SYMBOL,
            TokenType.NUMBER,
            TokenType.RIGHT_PAREN,
        ]

    def test_nested_sexp(self):
        scanner = Scanner("(foo (bar 42))")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert types == [
            TokenType.LEFT_PAREN,
            TokenType.SYMBOL,
            TokenType.LEFT_PAREN,
            TokenType.SYMBOL,
            TokenType.NUMBER,
            TokenType.RIGHT_PAREN,
            TokenType.RIGHT_PAREN,
        ]

    def test_sexp_with_string(self):
        scanner = Scanner('(key-sig "c major")')
        tokens = scanner.scan()
        string_token = [t for t in tokens if t.type == TokenType.STRING][0]
        assert string_token.literal == "c major"

    def test_sexp_with_negative_number(self):
        scanner = Scanner("(pan -50)")
        tokens = scanner.scan()
        num_token = [t for t in tokens if t.type == TokenType.NUMBER][0]
        assert num_token.literal == -50


class TestComments:
    """Test comment handling."""

    def test_comment_skipped(self):
        scanner = Scanner("c4 # this is a comment\nd4")
        tokens = scanner.scan()
        note_tokens = [t for t in tokens if t.type == TokenType.NOTE_LETTER]
        assert len(note_tokens) == 2
        assert [t.literal for t in note_tokens] == ["c", "d"]

    def test_comment_only(self):
        scanner = Scanner("# just a comment")
        tokens = scanner.scan()
        assert len(tokens) == 1  # Just EOF
        assert tokens[0].type == TokenType.EOF


class TestPositionTracking:
    """Test source position tracking."""

    def test_line_tracking(self):
        scanner = Scanner("c\nd\ne")
        tokens = scanner.scan()
        note_tokens = [t for t in tokens if t.type == TokenType.NOTE_LETTER]
        assert note_tokens[0].position.line == 1
        assert note_tokens[1].position.line == 2
        assert note_tokens[2].position.line == 3

    def test_column_tracking(self):
        scanner = Scanner("c d e")
        tokens = scanner.scan()
        note_tokens = [t for t in tokens if t.type == TokenType.NOTE_LETTER]
        assert note_tokens[0].position.column == 1
        assert note_tokens[1].position.column == 3
        assert note_tokens[2].position.column == 5


class TestErrors:
    """Test error handling."""

    def test_unexpected_character(self):
        scanner = Scanner("$")  # $ is not a valid Alda character
        with pytest.raises(AldaScanError) as exc_info:
            scanner.scan()
        assert "Unexpected character" in str(exc_info.value)

    def test_unterminated_string(self):
        scanner = Scanner('"unterminated')
        with pytest.raises(AldaScanError) as exc_info:
            scanner.scan()
        assert "Unterminated" in str(exc_info.value)


class TestComplexExamples:
    """Test complete Alda snippets."""

    def test_simple_melody(self):
        scanner = Scanner("o4 c4 d e f | g2 a b > c")
        tokens = scanner.scan()
        # Should produce valid tokens without errors
        assert any(t.type == TokenType.OCTAVE_SET for t in tokens)
        assert any(t.type == TokenType.BARLINE for t in tokens)
        assert any(t.type == TokenType.OCTAVE_UP for t in tokens)

    def test_chord(self):
        scanner = Scanner("c/e/g")
        tokens = scanner.scan()
        note_tokens = [t for t in tokens if t.type == TokenType.NOTE_LETTER]
        sep_tokens = [t for t in tokens if t.type == TokenType.SEPARATOR]
        assert len(note_tokens) == 3
        assert len(sep_tokens) == 2

    def test_note_with_accidentals(self):
        scanner = Scanner("c+ d- e_")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert types == [
            TokenType.NOTE_LETTER,
            TokenType.SHARP,
            TokenType.NOTE_LETTER,
            TokenType.FLAT,
            TokenType.NOTE_LETTER,
            TokenType.NATURAL,
        ]


class TestVariables:
    """Test scanning of variable-related tokens."""

    def test_equals_token(self):
        scanner = Scanner("foo = c d e")
        tokens = scanner.scan()
        types = [
            t.type
            for t in tokens
            if t.type != TokenType.EOF and t.type != TokenType.NEWLINE
        ]
        assert TokenType.EQUALS in types

    def test_variable_name(self):
        scanner = Scanner("myMotif = c d e")
        tokens = scanner.scan()
        name_tokens = [t for t in tokens if t.type == TokenType.NAME]
        assert len(name_tokens) == 1
        assert name_tokens[0].literal == "myMotif"


class TestMarkers:
    """Test scanning of marker tokens."""

    def test_marker(self):
        scanner = Scanner("%verse")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.MARKER
        assert tokens[0].literal == "verse"

    def test_at_marker(self):
        scanner = Scanner("@chorus")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.AT_MARKER
        assert tokens[0].literal == "chorus"

    def test_marker_with_hyphen(self):
        scanner = Scanner("%my-marker")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.MARKER
        assert tokens[0].literal == "my-marker"


class TestVoices:
    """Test scanning of voice markers."""

    def test_voice_marker(self):
        scanner = Scanner("V1:")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.VOICE_MARKER
        assert tokens[0].literal == 1

    def test_voice_zero(self):
        scanner = Scanner("V0:")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.VOICE_MARKER
        assert tokens[0].literal == 0

    def test_multiple_voices(self):
        scanner = Scanner("V1: c d V2: e f V0:")
        tokens = scanner.scan()
        voice_tokens = [t for t in tokens if t.type == TokenType.VOICE_MARKER]
        assert len(voice_tokens) == 3
        assert [t.literal for t in voice_tokens] == [1, 2, 0]


class TestCramBrackets:
    """Test scanning of cram and bracket tokens."""

    def test_cram_braces(self):
        scanner = Scanner("{c d e}")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert TokenType.CRAM_OPEN in types
        assert TokenType.CRAM_CLOSE in types

    def test_brackets(self):
        scanner = Scanner("[c d e]")
        tokens = scanner.scan()
        types = [t.type for t in tokens if t.type != TokenType.EOF]
        assert TokenType.EVENT_SEQ_OPEN in types
        assert TokenType.EVENT_SEQ_CLOSE in types


class TestRepeats:
    """Test scanning of repeat tokens."""

    def test_repeat(self):
        scanner = Scanner("*4")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.REPEAT
        assert tokens[0].literal == 4

    def test_repeat_with_sequence(self):
        scanner = Scanner("[c d e]*3")
        tokens = scanner.scan()
        repeat_tokens = [t for t in tokens if t.type == TokenType.REPEAT]
        assert len(repeat_tokens) == 1
        assert repeat_tokens[0].literal == 3


class TestRepetitions:
    """Test scanning of on-repetition tokens."""

    def test_simple_repetition(self):
        scanner = Scanner("'1")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.REPETITIONS
        assert tokens[0].literal == "1"

    def test_range_repetition(self):
        scanner = Scanner("'1-3")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.REPETITIONS
        assert tokens[0].literal == "1-3"

    def test_multiple_ranges(self):
        scanner = Scanner("'1-3,5,7-9")
        tokens = scanner.scan()
        assert tokens[0].type == TokenType.REPETITIONS
        assert tokens[0].literal == "1-3,5,7-9"



from __future__ import annotations

from contextlib import contextmanager
from contextvars import ContextVar
from typing import TYPE_CHECKING, Any, Generator

if TYPE_CHECKING:
    from prompt_toolkit.input.base import Input
    from prompt_toolkit.output.base import Output

    from .application import Application

__all__ = [
    "AppSession",
    "get_app_session",
    "get_app",
    "get_app_or_none",
    "set_app",
    "create_app_session",
    "create_app_session_from_tty",
]


class AppSession:
    """
    An AppSession is an interactive session, usually connected to one terminal.
    Within one such session, interaction with many applications can happen, one
    after the other.

    The input/output device is not supposed to change during one session.

    Warning: Always use the `create_app_session` function to create an
    instance, so that it gets activated correctly.

    :param input: Use this as a default input for all applications
        running in this session, unless an input is passed to the `Application`
        explicitly.
    :param output: Use this as a default output.
    """

    def __init__(
        self, input: Input | None = None, output: Output | None = None
    ) -> None:
        self._input = input
        self._output = output

        # The application will be set dynamically by the `set_app` context
        # manager. This is called in the application itself.
        self.app: Application[Any] | None = None

    def __repr__(self) -> str:
        return f"AppSession(app={self.app!r})"

    @property
    def input(self) -> Input:
        if self._input is None:
            from prompt_toolkit.input.defaults import create_input

            self._input = create_input()
        return self._input

    @property
    def output(self) -> Output:
        if self._output is None:
            from prompt_toolkit.output.defaults import create_output

            self._output = create_output()
        return self._output


_current_app_session: ContextVar[AppSession] = ContextVar(
    "_current_app_session", default=AppSession()
)


def get_app_session() -> AppSession:
    return _current_app_session.get()


def get_app() -> Application[Any]:
    """
    Get the current active (running) Application.
    An :class:`.Application` is active during the
    :meth:`.Application.run_async` call.

    We assume that there can only be one :class:`.Application` active at the
    same time. There is only one terminal window, with only one stdin and
    stdout. This makes the code significantly easier than passing around the
    :class:`.Application` everywhere.

    If no :class:`.Application` is running, then return by default a
    :class:`.DummyApplication`. For practical reasons, we prefer to not raise
    an exception. This way, we don't have to check all over the place whether
    an actual `Application` was returned.

    (For applications like pymux where we can have more than one `Application`,
    we'll use a work-around to handle that.)
    """
    session = _current_app_session.get()
    if session.app is not None:
        return session.app

    from .dummy import DummyApplication

    return DummyApplication()


def get_app_or_none() -> Application[Any] | None:
    """
    Get the current active (running) Application, or return `None` if no
    application is running.
    """
    session = _current_app_session.get()
    return session.app


@contextmanager
def set_app(app: Application[Any]) -> Generator[None, None, None]:
    """
    Context manager that sets the given :class:`.Application` active in an
    `AppSession`.

    This should only be called by the `Application` itself.
    The application will automatically be active while its running. If you want
    the application to be active in other threads/coroutines, where that's not
    the case, use `contextvars.copy_context()`, or use `Application.context` to
    run it in the appropriate context.
    """
    session = _current_app_session.get()

    previous_app = session.app
    session.app = app
    try:
        yield
    finally:
        session.app = previous_app


@contextmanager
def create_app_session(
    input: Input | None = None, output: Output | None = None
) -> Generator[AppSession, None, None]:
    """
    Create a separate AppSession.

    This is useful if there can be multiple individual ``AppSession``'s going
    on. Like in the case of a Telnet/SSH server.
    """
    # If no input/output is specified, fall back to the current input/output,
    # if there was one that was set/created for the current session.
    # (Note that we check `_input`/`_output` and not `input`/`output`. This is
    # because we don't want to accidentally create a new input/output objects
    # here and store it in the "parent" `AppSession`. Especially, when
    # combining pytest's `capsys` fixture and `create_app_session`, sys.stdin
    # and sys.stderr are patched for every test, so we don't want to leak
    # those outputs object across `AppSession`s.)
    if input is None:
        input = get_app_session()._input
    if output is None:
        output = get_app_session()._output

    # Create new `AppSession` and activate.
    session = AppSession(input=input, output=output)

    token = _current_app_session.set(session)
    try:
        yield session
    finally:
        _current_app_session.reset(token)


@contextmanager
def create_app_session_from_tty() -> Generator[AppSession, None, None]:
    """
    Create `AppSession` that always prefers the TTY input/output.

    Even if `sys.stdin` and `sys.stdout` are connected to input/output pipes,
    this will still use the terminal for interaction (because `sys.stderr` is
    still connected to the terminal).

    Usage::

        from prompt_toolkit.shortcuts import prompt

        with create_app_session_from_tty():
            prompt('>')
    """
    from prompt_toolkit.input.defaults import create_input
    from prompt_toolkit.output.defaults import create_output

    input = create_input(always_prefer_tty=True)
    output = create_output(always_prefer_tty=True)

    with create_app_session(input=input, output=output) as app_session:
        yield app_session



"""
All modern web browsers support these 140 color names.
Taken from: https://www.w3schools.com/colors/colors_names.asp
"""

from __future__ import annotations

__all__ = [
    "NAMED_COLORS",
]


NAMED_COLORS: dict[str, str] = {
    "AliceBlue": "#f0f8ff",
    "AntiqueWhite": "#faebd7",
    "Aqua": "#00ffff",
    "Aquamarine": "#7fffd4",
    "Azure": "#f0ffff",
    "Beige": "#f5f5dc",
    "Bisque": "#ffe4c4",
    "Black": "#000000",
    "BlanchedAlmond": "#ffebcd",
    "Blue": "#0000ff",
    "BlueViolet": "#8a2be2",
    "Brown": "#a52a2a",
    "BurlyWood": "#deb887",
    "CadetBlue": "#5f9ea0",
    "Chartreuse": "#7fff00",
    "Chocolate": "#d2691e",
    "Coral": "#ff7f50",
    "CornflowerBlue": "#6495ed",
    "Cornsilk": "#fff8dc",
    "Crimson": "#dc143c",
    "Cyan": "#00ffff",
    "DarkBlue": "#00008b",
    "DarkCyan": "#008b8b",
    "DarkGoldenRod": "#b8860b",
    "DarkGray": "#a9a9a9",
    "DarkGreen": "#006400",
    "DarkGrey": "#a9a9a9",
    "DarkKhaki": "#bdb76b",
    "DarkMagenta": "#8b008b",
    "DarkOliveGreen": "#556b2f",
    "DarkOrange": "#ff8c00",
    "DarkOrchid": "#9932cc",
    "DarkRed": "#8b0000",
    "DarkSalmon": "#e9967a",
    "DarkSeaGreen": "#8fbc8f",
    "DarkSlateBlue": "#483d8b",
    "DarkSlateGray": "#2f4f4f",
    "DarkSlateGrey": "#2f4f4f",
    "DarkTurquoise": "#00ced1",
    "DarkViolet": "#9400d3",
    "DeepPink": "#ff1493",
    "DeepSkyBlue": "#00bfff",
    "DimGray": "#696969",
    "DimGrey": "#696969",
    "DodgerBlue": "#1e90ff",
    "FireBrick": "#b22222",
    "FloralWhite": "#fffaf0",
    "ForestGreen": "#228b22",
    "Fuchsia": "#ff00ff",
    "Gainsboro": "#dcdcdc",
    "GhostWhite": "#f8f8ff",
    "Gold": "#ffd700",
    "GoldenRod": "#daa520",
    "Gray": "#808080",
    "Green": "#008000",
    "GreenYellow": "#adff2f",
    "Grey": "#808080",
    "HoneyDew": "#f0fff0",
    "HotPink": "#ff69b4",
    "IndianRed": "#cd5c5c",
    "Indigo": "#4b0082",
    "Ivory": "#fffff0",
    "Khaki": "#f0e68c",
    "Lavender": "#e6e6fa",
    "LavenderBlush": "#fff0f5",
    "LawnGreen": "#7cfc00",
    "LemonChiffon": "#fffacd",
    "LightBlue": "#add8e6",
    "LightCoral": "#f08080",
    "LightCyan": "#e0ffff",
    "LightGoldenRodYellow": "#fafad2",
    "LightGray": "#d3d3d3",
    "LightGreen": "#90ee90",
    "LightGrey": "#d3d3d3",
    "LightPink": "#ffb6c1",
    "LightSalmon": "#ffa07a",
    "LightSeaGreen": "#20b2aa",
    "LightSkyBlue": "#87cefa",
    "LightSlateGray": "#778899",
    "LightSlateGrey": "#778899",
    "LightSteelBlue": "#b0c4de",
    "LightYellow": "#ffffe0",
    "Lime": "#00ff00",
    "LimeGreen": "#32cd32",
    "Linen": "#faf0e6",
    "Magenta": "#ff00ff",
    "Maroon": "#800000",
    "MediumAquaMarine": "#66cdaa",
    "MediumBlue": "#0000cd",
    "MediumOrchid": "#ba55d3",
    "MediumPurple": "#9370db",
    "MediumSeaGreen": "#3cb371",
    "MediumSlateBlue": "#7b68ee",
    "MediumSpringGreen": "#00fa9a",
    "MediumTurquoise": "#48d1cc",
    "MediumVioletRed": "#c71585",
    "MidnightBlue": "#191970",
    "MintCream": "#f5fffa",
    "MistyRose": "#ffe4e1",
    "Moccasin": "#ffe4b5",
    "NavajoWhite": "#ffdead",
    "Navy": "#000080",
    "OldLace": "#fdf5e6",
    "Olive": "#808000",
    "OliveDrab": "#6b8e23",
    "Orange": "#ffa500",
    "OrangeRed": "#ff4500",
    "Orchid": "#da70d6",
    "PaleGoldenRod": "#eee8aa",
    "PaleGreen": "#98fb98",
    "PaleTurquoise": "#afeeee",
    "PaleVioletRed": "#db7093",
    "PapayaWhip": "#ffefd5",
    "PeachPuff": "#ffdab9",
    "Peru": "#cd853f",
    "Pink": "#ffc0cb",
    "Plum": "#dda0dd",
    "PowderBlue": "#b0e0e6",
    "Purple": "#800080",
    "RebeccaPurple": "#663399",
    "Red": "#ff0000",
    "RosyBrown": "#bc8f8f",
    "RoyalBlue": "#4169e1",
    "SaddleBrown": "#8b4513",
    "Salmon": "#fa8072",
    "SandyBrown": "#f4a460",
    "SeaGreen": "#2e8b57",
    "SeaShell": "#fff5ee",
    "Sienna": "#a0522d",
    "Silver": "#c0c0c0",
    "SkyBlue": "#87ceeb",
    "SlateBlue": "#6a5acd",
    "SlateGray": "#708090",
    "SlateGrey": "#708090",
    "Snow": "#fffafa",
    "SpringGreen": "#00ff7f",
    "SteelBlue": "#4682b4",
    "Tan": "#d2b48c",
    "Teal": "#008080",
    "Thistle": "#d8bfd8",
    "Tomato": "#ff6347",
    "Turquoise": "#40e0d0",
    "Violet": "#ee82ee",
    "Wheat": "#f5deb3",
    "White": "#ffffff",
    "WhiteSmoke": "#f5f5f5",
    "Yellow": "#ffff00",
    "YellowGreen": "#9acd32",
}



"""
Styling for prompt_toolkit applications.
"""

from __future__ import annotations

from .base import (ANSI_COLOR_NAMES, DEFAULT_ATTRS, Attrs, BaseStyle,
                   DummyStyle, DynamicStyle)
from .defaults import default_pygments_style, default_ui_style
from .named_colors import NAMED_COLORS
from .pygments import (pygments_token_to_classname, style_from_pygments_cls,
                       style_from_pygments_dict)
from .style import Priority, Style, merge_styles, parse_color
from .style_transformation import (AdjustBrightnessStyleTransformation,
                                   ConditionalStyleTransformation,
                                   DummyStyleTransformation,
                                   DynamicStyleTransformation,
                                   ReverseStyleTransformation,
                                   SetDefaultColorStyleTransformation,
                                   StyleTransformation,
                                   SwapLightAndDarkStyleTransformation,
                                   merge_style_transformations)

__all__ = [
    # Base.
    "Attrs",
    "DEFAULT_ATTRS",
    "ANSI_COLOR_NAMES",
    "BaseStyle",
    "DummyStyle",
    "DynamicStyle",
    # Defaults.
    "default_ui_style",
    "default_pygments_style",
    # Style.
    "Style",
    "Priority",
    "merge_styles",
    "parse_color",
    # Style transformation.
    "StyleTransformation",
    "SwapLightAndDarkStyleTransformation",
    "ReverseStyleTransformation",
    "SetDefaultColorStyleTransformation",
    "AdjustBrightnessStyleTransformation",
    "DummyStyleTransformation",
    "ConditionalStyleTransformation",
    "DynamicStyleTransformation",
    "merge_style_transformations",
    # Pygments.
    "style_from_pygments_cls",
    "style_from_pygments_dict",
    "pygments_token_to_classname",
    # Named colors.
    "NAMED_COLORS",
]



"""Tests for the compose.generate module - generative functions."""

import pytest
from aldakit.compose import (
    Note,
    Rest,
    note,
    rest,
    seq,
)
from aldakit.compose.generate import (
    # Random selection
    random_note,
    random_choice,
    weighted_choice,
    # Random walks
    random_walk,
    drunk_walk,
    # Rhythmic generators
    euclidean,
    probability_seq,
    rest_probability,
    # Markov chains
    MarkovChain,
    markov_chain,
    learn_markov,
    # L-Systems
    lsystem,
    # Cellular automata
    cellular_automaton,
    # Shift registers
    shift_register,
    turing_machine,
)


# =============================================================================
# Random Selection Tests
# =============================================================================


class TestRandomNote:
    def test_random_note_default_scale(self):
        """Random note from default C major scale."""
        n = random_note(seed=42)
        assert isinstance(n, Note)
        assert n.pitch in ["c", "d", "e", "f", "g", "a", "b"]

    def test_random_note_custom_scale(self):
        """Random note from custom scale."""
        scale = ["c", "e", "g"]
        n = random_note(scale=scale, seed=42)
        assert n.pitch in scale

    def test_random_note_with_duration(self):
        """Random note with specified duration."""
        n = random_note(duration=8, seed=42)
        assert n.duration == 8

    def test_random_note_with_octave(self):
        """Random note with specified octave."""
        n = random_note(octave=5, seed=42)
        assert n.octave == 5

    def test_random_note_reproducible_with_seed(self):
        """Same seed produces same note."""
        n1 = random_note(seed=123)
        n2 = random_note(seed=123)
        assert n1.pitch == n2.pitch


class TestRandomChoice:
    def test_random_choice_basic(self):
        """Choose from list of notes."""
        options = [note("c"), note("e"), note("g")]
        result = random_choice(options, seed=42)
        assert result in options

    def test_random_choice_reproducible(self):
        """Same seed produces same choice."""
        options = [note("c"), note("d"), note("e"), note("f")]
        r1 = random_choice(options, seed=99)
        r2 = random_choice(options, seed=99)
        assert r1.pitch == r2.pitch

    def test_random_choice_empty_raises(self):
        """Empty list raises error."""
        with pytest.raises(ValueError, match="cannot be empty"):
            random_choice([])


class TestWeightedChoice:
    def test_weighted_choice_basic(self):
        """Weighted choice returns valid item."""
        options = [(note("c"), 0.5), (note("e"), 0.3), (note("g"), 0.2)]
        result = weighted_choice(options, seed=42)
        assert result.pitch in ["c", "e", "g"]

    def test_weighted_choice_heavy_weight(self):
        """Heavy weight dominates selection (statistical test)."""
        options = [(note("c"), 100), (note("e"), 1)]
        # Run multiple times - should mostly get "c"
        results = [weighted_choice(options, seed=i).pitch for i in range(100)]
        c_count = results.count("c")
        assert c_count > 90  # Should be heavily biased toward "c"

    def test_weighted_choice_empty_raises(self):
        """Empty list raises error."""
        with pytest.raises(ValueError, match="cannot be empty"):
            weighted_choice([])

    def test_weighted_choice_negative_weight_raises(self):
        """Negative weight raises error."""
        with pytest.raises(ValueError, match="cannot be negative"):
            weighted_choice([(note("c"), -1)])

    def test_weighted_choice_zero_total_raises(self):
        """Zero total weight raises error."""
        with pytest.raises(ValueError, match="cannot be zero"):
            weighted_choice([(note("c"), 0), (note("e"), 0)])


# =============================================================================
# Random Walk Tests
# =============================================================================


class TestRandomWalk:
    def test_random_walk_basic(self):
        """Basic random walk produces sequence."""
        result = random_walk("c", 8, seed=42)
        assert len(result.elements) == 8
        assert all(isinstance(e, Note) for e in result.elements)

    def test_random_walk_starts_at_pitch(self):
        """First note is the start pitch."""
        result = random_walk("e", 5, octave=5, seed=42)
        assert result.elements[0].pitch == "e"
        assert result.elements[0].octave == 5

    def test_random_walk_with_duration(self):
        """All notes have specified duration."""
        result = random_walk("c", 5, duration=8, seed=42)
        assert all(n.duration == 8 for n in result.elements)

    def test_random_walk_stays_in_range(self):
        """Notes stay within pitch range."""
        result = random_walk("c", 50, min_pitch=48, max_pitch=72, seed=42)
        for n in result.elements:
            assert 48 <= n.midi_pitch <= 72

    def test_random_walk_custom_intervals(self):
        """Custom intervals are used."""
        result = random_walk("c", 10, intervals=[1, 2], seed=42)
        # All intervals should be positive, so pitch should trend up
        assert len(result.elements) == 10

    def test_random_walk_zero_steps(self):
        """Zero steps returns empty sequence."""
        result = random_walk("c", 0, seed=42)
        assert len(result.elements) == 0

    def test_random_walk_reproducible(self):
        """Same seed produces same walk."""
        r1 = random_walk("c", 10, seed=123)
        r2 = random_walk("c", 10, seed=123)
        for n1, n2 in zip(r1.elements, r2.elements):
            assert n1.pitch == n2.pitch


class TestDrunkWalk:
    def test_drunk_walk_basic(self):
        """Basic drunk walk produces sequence."""
        result = drunk_walk("c", 8, seed=42)
        assert len(result.elements) == 8

    def test_drunk_walk_small_steps(self):
        """Max step constrains interval size."""
        result = drunk_walk("c", 20, max_step=1, octave=4, seed=42)
        # With max_step=1, intervals should be small
        assert len(result.elements) == 20

    def test_drunk_walk_with_bias(self):
        """Bias affects direction tendency."""
        # Positive bias should trend upward
        result = drunk_walk("c", 20, max_step=2, bias=0.5, octave=4, seed=42)
        assert len(result.elements) == 20

    def test_drunk_walk_reproducible(self):
        """Same seed produces same walk."""
        r1 = drunk_walk("e", 8, seed=99)
        r2 = drunk_walk("e", 8, seed=99)
        for n1, n2 in zip(r1.elements, r2.elements):
            assert n1.pitch == n2.pitch


# =============================================================================
# Rhythmic Generator Tests
# =============================================================================


class TestEuclidean:
    def test_euclidean_basic(self):
        """Basic Euclidean rhythm."""
        result = euclidean(3, 8, "c")
        assert len(result.elements) == 8
        hits = sum(1 for e in result.elements if isinstance(e, Note))
        assert hits == 3

    def test_euclidean_tresillo(self):
        """3 over 8 is the Cuban tresillo pattern."""
        result = euclidean(3, 8, "c")
        pattern = [isinstance(e, Note) for e in result.elements]
        # Should be [T, F, F, T, F, F, T, F] or similar
        assert sum(pattern) == 3

    def test_euclidean_cinquillo(self):
        """5 over 8 is the cinquillo pattern."""
        result = euclidean(5, 8, "c")
        hits = sum(1 for e in result.elements if isinstance(e, Note))
        assert hits == 5

    def test_euclidean_with_duration(self):
        """Notes and rests have specified duration."""
        result = euclidean(3, 8, "c", duration=16)
        for elem in result.elements:
            if isinstance(elem, Note):
                assert elem.duration == 16
            else:
                assert elem.duration == 16

    def test_euclidean_rotation(self):
        """Rotation shifts the pattern."""
        r1 = euclidean(3, 8, "c")
        r2 = euclidean(3, 8, "c", rotate=1)
        # Rotated pattern should be different
        p1 = [isinstance(e, Note) for e in r1.elements]
        p2 = [isinstance(e, Note) for e in r2.elements]
        assert p1 != p2
        # But same number of hits
        assert sum(p1) == sum(p2)

    def test_euclidean_all_hits(self):
        """All hits (hits == steps)."""
        result = euclidean(8, 8, "c")
        assert all(isinstance(e, Note) for e in result.elements)

    def test_euclidean_no_hits(self):
        """No hits (hits == 0)."""
        result = euclidean(0, 8, "c")
        assert all(isinstance(e, Rest) for e in result.elements)

    def test_euclidean_empty(self):
        """Zero steps returns empty."""
        result = euclidean(0, 0, "c")
        assert len(result.elements) == 0

    def test_euclidean_invalid_hits_raises(self):
        """Hits > steps raises error."""
        with pytest.raises(ValueError, match="cannot be greater"):
            euclidean(10, 5, "c")


class TestProbabilitySeq:
    def test_probability_seq_basic(self):
        """Basic probability sequence."""
        result = probability_seq(["c", "d", "e"], 10, probability=0.7, seed=42)
        assert len(result.elements) == 10

    def test_probability_seq_all_notes(self):
        """Probability 1.0 gives all notes."""
        result = probability_seq(["c"], 10, probability=1.0, seed=42)
        assert all(isinstance(e, Note) for e in result.elements)

    def test_probability_seq_all_rests(self):
        """Probability 0.0 gives all rests."""
        result = probability_seq(["c"], 10, probability=0.0, seed=42)
        assert all(isinstance(e, Rest) for e in result.elements)

    def test_probability_seq_with_duration(self):
        """Notes have specified duration."""
        result = probability_seq(["c"], 5, probability=1.0, duration=8, seed=42)
        for elem in result.elements:
            if isinstance(elem, Note):
                assert elem.duration == 8

    def test_probability_seq_empty_notes_raises(self):
        """Empty notes list raises error."""
        with pytest.raises(ValueError, match="cannot be empty"):
            probability_seq([], 10)

    def test_probability_seq_invalid_probability_raises(self):
        """Invalid probability raises error."""
        with pytest.raises(ValueError, match="between 0 and 1"):
            probability_seq(["c"], 10, probability=1.5)


class TestRestProbability:
    def test_rest_probability_basic(self):
        """Basic rest probability."""
        melody = seq(note("c"), note("d"), note("e"), note("f"))
        result = rest_probability(melody, 0.5, seed=42)
        assert len(result.elements) == 4

    def test_rest_probability_zero(self):
        """Probability 0 keeps all notes."""
        melody = seq(note("c"), note("d"), note("e"))
        result = rest_probability(melody, 0.0, seed=42)
        assert all(isinstance(e, Note) for e in result.elements)

    def test_rest_probability_one(self):
        """Probability 1 replaces all notes."""
        melody = seq(note("c"), note("d"), note("e"))
        result = rest_probability(melody, 1.0, seed=42)
        assert all(isinstance(e, Rest) for e in result.elements)

    def test_rest_probability_preserves_rests(self):
        """Existing rests are preserved."""
        melody = seq(note("c"), rest(), note("e"))
        result = rest_probability(melody, 0.5, seed=42)
        # The original rest should still be there
        assert any(isinstance(e, Rest) for e in result.elements)


# =============================================================================
# Markov Chain Tests
# =============================================================================


class TestMarkovChain:
    def test_markov_chain_basic(self):
        """Basic Markov chain generation."""
        chain = markov_chain(
            {
                "c": {"d": 0.5, "e": 0.5},
                "d": {"e": 0.5, "c": 0.5},
                "e": {"c": 1.0},
            }
        )
        result = chain.generate(start="c", length=10, seed=42)
        assert len(result.elements) == 10
        assert all(isinstance(e, Note) for e in result.elements)

    def test_markov_chain_starts_correct(self):
        """Chain starts with specified note."""
        chain = markov_chain({"c": {"d": 1.0}, "d": {"c": 1.0}})
        result = chain.generate(start="c", length=5, seed=42)
        assert result.elements[0].pitch == "c"

    def test_markov_chain_follows_transitions(self):
        """Chain follows deterministic transitions."""
        chain = markov_chain(
            {
                "c": {"d": 1.0},
                "d": {"e": 1.0},
                "e": {"c": 1.0},
            }
        )
        result = chain.generate(start="c", length=6, seed=42)
        pitches = [n.pitch for n in result.elements]
        assert pitches == ["c", "d", "e", "c", "d", "e"]

    def test_markov_chain_with_duration(self):
        """Generated notes have specified duration."""
        chain = markov_chain({"c": {"d": 1.0}, "d": {"c": 1.0}})
        result = chain.generate(start="c", length=5, duration=8, seed=42)
        assert all(n.duration == 8 for n in result.elements)

    def test_markov_chain_reproducible(self):
        """Same seed produces same sequence."""
        chain = markov_chain(
            {
                "c": {"d": 0.5, "e": 0.5},
                "d": {"c": 0.5, "e": 0.5},
                "e": {"c": 0.5, "d": 0.5},
            }
        )
        r1 = chain.generate(start="c", length=10, seed=123)
        r2 = chain.generate(start="c", length=10, seed=123)
        for n1, n2 in zip(r1.elements, r2.elements):
            assert n1.pitch == n2.pitch

    def test_markov_chain_empty_raises(self):
        """Empty chain raises error."""
        chain = MarkovChain()
        with pytest.raises(ValueError, match="no transitions"):
            chain.generate(length=5)


class TestLearnMarkov:
    def test_learn_markov_basic(self):
        """Learn from simple sequence."""
        melody = seq(note("c"), note("d"), note("e"), note("d"), note("c"))
        chain = learn_markov(melody)
        # Should be able to generate
        result = chain.generate(length=10, seed=42)
        assert len(result.elements) == 10

    def test_learn_markov_learns_transitions(self):
        """Learned transitions reflect input."""
        # c always goes to d, d always goes to c
        melody = seq(note("c"), note("d"), note("c"), note("d"))
        chain = learn_markov(melody)
        assert "c" in chain.transitions
        assert "d" in chain.transitions

    def test_learn_markov_higher_order(self):
        """Higher-order Markov chain."""
        melody = seq(
            note("c"),
            note("d"),
            note("e"),
            note("c"),
            note("d"),
            note("f"),
        )
        chain = learn_markov(melody, order=2)
        assert chain.order == 2

    def test_learn_markov_short_sequence_raises(self):
        """Sequence too short for order raises error."""
        melody = seq(note("c"))
        with pytest.raises(ValueError, match="too short"):
            learn_markov(melody, order=1)


# =============================================================================
# L-System Tests
# =============================================================================


class TestLSystem:
    def test_lsystem_basic(self):
        """Basic L-system generation."""
        result = lsystem(
            axiom="A",
            rules={"A": "AB", "B": "A"},
            iterations=3,
            note_map={"A": note("c"), "B": note("e")},
        )
        assert len(result.elements) > 0
        assert all(isinstance(e, Note) for e in result.elements)

    def test_lsystem_fibonacci_length(self):
        """Fibonacci L-system produces correct length."""
        # A -> AB, B -> A produces Fibonacci sequence
        # iter 0: A (1)
        # iter 1: AB (2)
        # iter 2: ABA (3)
        # iter 3: ABAAB (5)
        # iter 4: ABAABABA (8)
        result = lsystem(
            axiom="A",
            rules={"A": "AB", "B": "A"},
            iterations=4,
            note_map={"A": note("c"), "B": note("e")},
        )
        assert len(result.elements) == 8  # Fibonacci number

    def test_lsystem_no_iterations(self):
        """Zero iterations returns axiom only."""
        result = lsystem(
            axiom="ABC",
            rules={"A": "XY"},
            iterations=0,
            note_map={"A": note("c"), "B": note("d"), "C": note("e")},
        )
        assert len(result.elements) == 3

    def test_lsystem_unmapped_symbols_ignored(self):
        """Unmapped symbols are ignored by default."""
        result = lsystem(
            axiom="A+B",
            rules={"A": "A", "B": "B"},
            iterations=1,
            note_map={"A": note("c"), "B": note("e")},
        )
        # Only A and B are mapped, + is ignored
        assert len(result.elements) == 2

    def test_lsystem_with_default(self):
        """Unmapped symbols use default."""
        result = lsystem(
            axiom="A+B",
            rules={},
            iterations=0,
            note_map={"A": note("c"), "B": note("e")},
            default=note("g"),
        )
        # A, +, B all become notes
        assert len(result.elements) == 3

    def test_lsystem_with_rests(self):
        """L-system can include rests."""
        result = lsystem(
            axiom="AB",
            rules={"A": "AB", "B": "A"},
            iterations=2,
            note_map={"A": note("c"), "B": rest()},
        )
        notes = [e for e in result.elements if isinstance(e, Note)]
        rests = [e for e in result.elements if isinstance(e, Rest)]
        assert len(notes) > 0
        assert len(rests) > 0


# =============================================================================
# Cellular Automaton Tests
# =============================================================================


class TestCellularAutomaton:
    def test_cellular_automaton_basic(self):
        """Basic cellular automaton generation."""
        result = cellular_automaton(rule=110, width=8, steps=4, pitch_on="c")
        assert len(result.elements) == 8 * 4  # width * steps

    def test_cellular_automaton_rule_90(self):
        """Rule 90 (Sierpinski triangle) produces pattern."""
        result = cellular_automaton(rule=90, width=8, steps=4, pitch_on="c")
        # Should have mix of notes and rests
        notes = sum(1 for e in result.elements if isinstance(e, Note))
        rests = sum(1 for e in result.elements if isinstance(e, Rest))
        assert notes > 0
        assert rests > 0

    def test_cellular_automaton_with_duration(self):
        """Notes have specified duration."""
        result = cellular_automaton(
            rule=30, width=4, steps=2, pitch_on="c", duration=16
        )
        for elem in result.elements:
            if isinstance(elem, Note):
                assert elem.duration == 16

    def test_cellular_automaton_custom_initial(self):
        """Custom initial state."""
        result = cellular_automaton(
            rule=110,
            width=4,
            steps=1,
            pitch_on="c",
            initial=[True, False, True, False],
        )
        # First row should match initial pattern
        pattern = [isinstance(e, Note) for e in result.elements[:4]]
        assert pattern == [True, False, True, False]

    def test_cellular_automaton_invalid_rule_raises(self):
        """Invalid rule number raises error."""
        with pytest.raises(ValueError, match="between 0 and 255"):
            cellular_automaton(rule=300, width=8, steps=4, pitch_on="c")

    def test_cellular_automaton_invalid_initial_raises(self):
        """Wrong initial length raises error."""
        with pytest.raises(ValueError, match="must have length"):
            cellular_automaton(
                rule=110,
                width=8,
                steps=4,
                pitch_on="c",
                initial=[True, False],  # Wrong length
            )


# =============================================================================
# Shift Register Tests
# =============================================================================


class TestShiftRegister:
    def test_shift_register_basic(self):
        """Basic LFSR generation."""
        result = shift_register(16, bits=8)
        assert len(result.elements) == 16
        assert all(isinstance(e, Note) for e in result.elements)

    def test_shift_register_with_scale(self):
        """LFSR with custom scale."""
        scale = ["c", "e", "g"]
        result = shift_register(10, bits=4, scale=scale)
        for elem in result.elements:
            assert elem.pitch in scale

    def test_shift_register_with_duration(self):
        """LFSR with specified duration."""
        result = shift_register(8, bits=4, duration=8)
        assert all(n.duration == 8 for n in result.elements)

    def test_shift_register_deterministic(self):
        """Same initial value produces same sequence."""
        r1 = shift_register(16, bits=8, initial=42)
        r2 = shift_register(16, bits=8, initial=42)
        for n1, n2 in zip(r1.elements, r2.elements):
            assert n1.pitch == n2.pitch

    def test_shift_register_binary_mode(self):
        """Binary mode produces notes and rests."""
        result = shift_register(16, bits=8, mode="binary")
        notes = [e for e in result.elements if isinstance(e, Note)]
        rests = [e for e in result.elements if isinstance(e, Rest)]
        # Should have some of each (statistically likely)
        assert len(notes) + len(rests) == 16

    def test_shift_register_custom_taps(self):
        """Custom tap positions work."""
        result = shift_register(16, taps=[0, 3], bits=4)
        assert len(result.elements) == 16

    def test_shift_register_cycles(self):
        """LFSR with maximal taps should cycle after 2^n - 1 steps."""
        # 4-bit LFSR has max cycle of 15
        result = shift_register(30, bits=4, initial=1)
        pitches = [n.pitch for n in result.elements]
        # First 15 should equal second 15
        assert pitches[:15] == pitches[15:]

    def test_shift_register_invalid_bits_raises(self):
        """Invalid bit size raises error."""
        with pytest.raises(ValueError, match="between 1 and 16"):
            shift_register(16, bits=0)
        with pytest.raises(ValueError, match="between 1 and 16"):
            shift_register(16, bits=20)

    def test_shift_register_invalid_tap_raises(self):
        """Tap out of range raises error."""
        with pytest.raises(ValueError, match="out of range"):
            shift_register(16, taps=[0, 10], bits=4)


class TestTuringMachine:
    def test_turing_machine_basic(self):
        """Basic Turing Machine generation."""
        result = turing_machine(16, bits=8, seed=42)
        assert len(result.elements) == 16
        assert all(isinstance(e, Note) for e in result.elements)

    def test_turing_machine_locked_loop(self):
        """Probability 0 creates exact loop."""
        result = turing_machine(24, bits=8, probability=0.0, seed=42)
        pitches = [n.pitch for n in result.elements]
        # Should repeat after 8 steps
        assert pitches[:8] == pitches[8:16]
        assert pitches[:8] == pitches[16:24]

    def test_turing_machine_with_scale(self):
        """Turing Machine with custom scale."""
        scale = ["c", "d", "e", "g", "a"]
        result = turing_machine(16, bits=8, scale=scale, seed=42)
        for elem in result.elements:
            assert elem.pitch in scale

    def test_turing_machine_evolving(self):
        """High probability causes variation."""
        result = turing_machine(32, bits=8, probability=0.5, seed=42)
        pitches = [n.pitch for n in result.elements]
        # With 50% flip probability, first 8 unlikely to equal second 8
        # (This is a statistical test, could theoretically fail)
        first_8 = pitches[:8]
        second_8 = pitches[8:16]
        third_8 = pitches[16:24]
        # At least one should be different
        assert first_8 != second_8 or second_8 != third_8

    def test_turing_machine_reproducible(self):
        """Same seed produces same sequence."""
        r1 = turing_machine(16, bits=8, probability=0.3, seed=123)
        r2 = turing_machine(16, bits=8, probability=0.3, seed=123)
        for n1, n2 in zip(r1.elements, r2.elements):
            assert n1.pitch == n2.pitch

    def test_turing_machine_with_duration(self):
        """Turing Machine with specified duration."""
        result = turing_machine(8, bits=4, duration=16, seed=42)
        assert all(n.duration == 16 for n in result.elements)

    def test_turing_machine_invalid_probability_raises(self):
        """Invalid probability raises error."""
        with pytest.raises(ValueError, match="between 0 and 1"):
            turing_machine(16, probability=1.5)


# =============================================================================
# Integration Tests
# =============================================================================


class TestIntegration:
    def test_euclidean_to_alda(self):
        """Euclidean rhythm can be exported to Alda."""
        result = euclidean(3, 8, "c", duration=8)
        alda = result.to_alda()
        assert "c8" in alda or "r8" in alda

    def test_random_walk_to_alda(self):
        """Random walk can be exported to Alda."""
        result = random_walk("c", 4, duration=8, seed=42)
        alda = result.to_alda()
        assert len(alda) > 0

    def test_markov_to_alda(self):
        """Markov chain output can be exported to Alda."""
        chain = markov_chain({"c": {"d": 1.0}, "d": {"c": 1.0}})
        result = chain.generate(start="c", length=4, duration=8, seed=42)
        alda = result.to_alda()
        assert "c8" in alda or "d8" in alda

    def test_combine_with_score(self):
        """Generative functions work with Score."""
        from aldakit import Score
        from aldakit.compose import part, tempo

        rhythm = euclidean(3, 8, "c", duration=8)
        walk = random_walk("e", 8, duration=8, seed=42)

        score = Score.from_elements(
            part("piano"),
            tempo(120),
            *rhythm.elements,
            *walk.elements,
        )

        assert score.duration > 0
        assert len(score.ast.children) > 0

    def test_chain_generators(self):
        """Chain multiple generators together."""
        # Generate a melody with random walk
        melody = random_walk("c", 8, duration=8, seed=42)

        # Add random rests
        sparse = rest_probability(melody, 0.3, seed=42)

        # Should still be valid
        assert len(sparse.elements) == 8
        alda = sparse.to_alda()
        assert len(alda) > 0



"""
patch_stdout
============

This implements a context manager that ensures that print statements within
it won't destroy the user interface. The context manager will replace
`sys.stdout` by something that draws the output above the current prompt,
rather than overwriting the UI.

Usage::

    with patch_stdout(application):
        ...
        application.run()
        ...

Multiple applications can run in the body of the context manager, one after the
other.
"""

from __future__ import annotations

import asyncio
import queue
import sys
import threading
import time
from contextlib import contextmanager
from typing import Generator, TextIO, cast

from .application import get_app_session, run_in_terminal
from .output import Output

__all__ = [
    "patch_stdout",
    "StdoutProxy",
]


@contextmanager
def patch_stdout(raw: bool = False) -> Generator[None, None, None]:
    """
    Replace `sys.stdout` by an :class:`_StdoutProxy` instance.

    Writing to this proxy will make sure that the text appears above the
    prompt, and that it doesn't destroy the output from the renderer.  If no
    application is curring, the behavior should be identical to writing to
    `sys.stdout` directly.

    Warning: If a new event loop is installed using `asyncio.set_event_loop()`,
        then make sure that the context manager is applied after the event loop
        is changed. Printing to stdout will be scheduled in the event loop
        that's active when the context manager is created.

    :param raw: (`bool`) When True, vt100 terminal escape sequences are not
                removed/escaped.
    """
    with StdoutProxy(raw=raw) as proxy:
        original_stdout = sys.stdout
        original_stderr = sys.stderr

        # Enter.
        sys.stdout = cast(TextIO, proxy)
        sys.stderr = cast(TextIO, proxy)

        try:
            yield
        finally:
            sys.stdout = original_stdout
            sys.stderr = original_stderr


class _Done:
    "Sentinel value for stopping the stdout proxy."


class StdoutProxy:
    """
    File-like object, which prints everything written to it, output above the
    current application/prompt. This class is compatible with other file
    objects and can be used as a drop-in replacement for `sys.stdout` or can
    for instance be passed to `logging.StreamHandler`.

    The current application, above which we print, is determined by looking
    what application currently runs in the `AppSession` that is active during
    the creation of this instance.

    This class can be used as a context manager.

    In order to avoid having to repaint the prompt continuously for every
    little write, a short delay of `sleep_between_writes` seconds will be added
    between writes in order to bundle many smaller writes in a short timespan.
    """

    def __init__(
        self,
        sleep_between_writes: float = 0.2,
        raw: bool = False,
    ) -> None:
        self.sleep_between_writes = sleep_between_writes
        self.raw = raw

        self._lock = threading.RLock()
        self._buffer: list[str] = []

        # Keep track of the curret app session.
        self.app_session = get_app_session()

        # See what output is active *right now*. We should do it at this point,
        # before this `StdoutProxy` instance is possibly assigned to `sys.stdout`.
        # Otherwise, if `patch_stdout` is used, and no `Output` instance has
        # been created, then the default output creation code will see this
        # proxy object as `sys.stdout`, and get in a recursive loop trying to
        # access `StdoutProxy.isatty()` which will again retrieve the output.
        self._output: Output = self.app_session.output

        # Flush thread
        self._flush_queue: queue.Queue[str | _Done] = queue.Queue()
        self._flush_thread = self._start_write_thread()
        self.closed = False

    def __enter__(self) -> StdoutProxy:
        return self

    def __exit__(self, *args: object) -> None:
        self.close()

    def close(self) -> None:
        """
        Stop `StdoutProxy` proxy.

        This will terminate the write thread, make sure everything is flushed
        and wait for the write thread to finish.
        """
        if not self.closed:
            self._flush_queue.put(_Done())
            self._flush_thread.join()
            self.closed = True

    def _start_write_thread(self) -> threading.Thread:
        thread = threading.Thread(
            target=self._write_thread,
            name="patch-stdout-flush-thread",
            daemon=True,
        )
        thread.start()
        return thread

    def _write_thread(self) -> None:
        done = False

        while not done:
            item = self._flush_queue.get()

            if isinstance(item, _Done):
                break

            # Don't bother calling when we got an empty string.
            if not item:
                continue

            text = []
            text.append(item)

            # Read the rest of the queue if more data was queued up.
            while True:
                try:
                    item = self._flush_queue.get_nowait()
                except queue.Empty:
                    break
                else:
                    if isinstance(item, _Done):
                        done = True
                    else:
                        text.append(item)

            app_loop = self._get_app_loop()
            self._write_and_flush(app_loop, "".join(text))

            # If an application was running that requires repainting, then wait
            # for a very short time, in order to bundle actual writes and avoid
            # having to repaint to often.
            if app_loop is not None:
                time.sleep(self.sleep_between_writes)

    def _get_app_loop(self) -> asyncio.AbstractEventLoop | None:
        """
        Return the event loop for the application currently running in our
        `AppSession`.
        """
        app = self.app_session.app

        if app is None:
            return None

        return app.loop

    def _write_and_flush(
        self, loop: asyncio.AbstractEventLoop | None, text: str
    ) -> None:
        """
        Write the given text to stdout and flush.
        If an application is running, use `run_in_terminal`.
        """

        def write_and_flush() -> None:
            # Ensure that autowrap is enabled before calling `write`.
            # XXX: On Windows, the `Windows10_Output` enables/disables VT
            #      terminal processing for every flush. It turns out that this
            #      causes autowrap to be reset (disabled) after each flush. So,
            #      we have to enable it again before writing text.
            self._output.enable_autowrap()

            if self.raw:
                self._output.write_raw(text)
            else:
                self._output.write(text)

            self._output.flush()

        def write_and_flush_in_loop() -> None:
            # If an application is running, use `run_in_terminal`, otherwise
            # call it directly.
            run_in_terminal(write_and_flush, in_executor=False)

        if loop is None:
            # No loop, write immediately.
            write_and_flush()
        else:
            # Make sure `write_and_flush` is executed *in* the event loop, not
            # in another thread.
            loop.call_soon_threadsafe(write_and_flush_in_loop)

    def _write(self, data: str) -> None:
        """
        Note: print()-statements cause to multiple write calls.
              (write('line') and write('\n')). Of course we don't want to call
              `run_in_terminal` for every individual call, because that's too
              expensive, and as long as the newline hasn't been written, the
              text itself is again overwritten by the rendering of the input
              command line. Therefor, we have a little buffer which holds the
              text until a newline is written to stdout.
        """
        if "\n" in data:
            # When there is a newline in the data, write everything before the
            # newline, including the newline itself.
            before, after = data.rsplit("\n", 1)
            to_write = self._buffer + [before, "\n"]
            self._buffer = [after]

            text = "".join(to_write)
            self._flush_queue.put(text)
        else:
            # Otherwise, cache in buffer.
            self._buffer.append(data)

    def _flush(self) -> None:
        text = "".join(self._buffer)
        self._buffer = []
        self._flush_queue.put(text)

    def write(self, data: str) -> int:
        with self._lock:
            self._write(data)

        return len(data)  # Pretend everything was written.

    def flush(self) -> None:
        """
        Flush buffered output.
        """
        with self._lock:
            self._flush()

    @property
    def original_stdout(self) -> TextIO | None:
        return self._output.stdout or sys.__stdout__

    # Attributes for compatibility with sys.__stdout__:

    def fileno(self) -> int:
        return self._output.fileno()

    def isatty(self) -> bool:
        stdout = self._output.stdout
        if stdout is None:
            return False

        return stdout.isatty()

    @property
    def encoding(self) -> str:
        return self._output.encoding()

    @property
    def errors(self) -> str:
        return "strict"
